/*
   $Id: cr2dump.cpp,v 1.1 2005/03/24 05:41:04 dmg Exp $

   Library to read the  TIFF data from Canon files

 * Copyright (C) 2004,2005 Daniel M. German
 *
 * Author: Daniel M. German dmg _at_ turingmachine _dot_ org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*

#ifndef __LIBCR2_H__
#define __LIBCR2_H__
#endif

Here is the design idea:

* An object for the photo. The photo will be responsible to load itself. 

* It should know how to parse itself.

* There will be also a Parser object. The parser will be responsible
* for doing the scanning of the records.



* Start with the photo object.
* A photo needs to load itself (calling parser)
* A photo should also have a method to extract its EXIF only
* BUt also to copy itself to another place


TODO

Error handling in all freads



*/

#define DEBUG


#include <string.h>
#include <stdio.h>
#include <assert.h>

#define USAGE "THis program is to be used ....\n"



#include "crawdebug.h"
#include "libcraw.h"


int debug = 0;

/*

 I will create a tree of entries.

A directory entry will be represented as one of these, with a count
equal to the number of entries in the directory. 
pChild will point to the array of these.
pNext will point to the next record

a dummy info_entry will represent the first directory entry in the file


*/

typedef struct {
   int id;
   char *name;
} lookup_table_entry;


#define JPEG_SOF0  0xC0
#define JPEG_SOF1  0xC1
#define JPEG_SOF2  0xC2
#define JPEG_SOF3  0xC3
#define JPEG_SOF4  0xC4
#define JPEG_SOF5  0xC5
#define JPEG_SOF6  0xC6
#define JPEG_SOF7  0xC7
#define JPEG_SOF8  0xC8
#define JPEG_SOF9  0xC9
#define JPEG_SOF10 0xCA
#define JPEG_SOF11 0xCB
#define JPEG_SOF12 0xCC
#define JPEG_SOF13 0xCD
#define JPEG_SOF14 0xCE
#define JPEG_SOF15 0xCF
#define JPEG_SOI   0xD8
#define JPEG_EOI   0xD9
#define JPEG_SOS   0xDA
#define JPEG_JFIF  0xE0
#define JPEG_EXIF  0xE1
#define JPEG_COM   0xFE


#define TIFF_EXIF_TAG             0x8769
#define TIFF_INTEROPERABILITY_TAG 0xa005
#define TIFF_MAKER_NOTE_TAG       0x927c


lookup_table_entry tiffTagsNames [] = {
   {0x0000,  "Canon tag 0 Always 0"},
   {0x0001,  "Canon tag 1 Photo Information (0x10d2 in raw first )"},
   {0x0002,  "Canon tag 2 Focal Length (0x5029 in raw)"},
   {0x0003,  "Canon tag 3 Unknown 1 (0x5028 in raw)"},
   {0x0004,  "Canon tag 4 Exposure information (0x102a in raw)"},
   {0x0006,  "Canon tag 6 Image name (0x0815 in raw)"},
   {0x0007,  "Canon tag 7 Camera Firmware (0x080b in raw)"},
   {0x0008,  "Canon tag 8  Photo serial number  (0x5817 in raw)"},
   {0x0009,  "Canon tag 9 Owner name (0x0810 in raw)"},
   {0x000c,  "Canon tag c Camera serial number (0x580b in raw)"},
   {0x000d,  "Canon tag d Image description??? (0x0805)"},
   {0x000e,  "Canon tag e"},
   {0x000f,  "Canon tag f Custom FNs           (0x1033 in raw)"},
   {0x0010,  "Canon tag 10"},
   {0x00a8,  "Canon tag a8 Unknown 15 (0x10a8 in raw)"},
   {0x00a9,  "Canon tag a9 White balance table (0x10a9 in raw)"},
   {0x00aa,  "Canon tag aa Unknown 14 (0x10aa in raw)"},
   {0x00c0,  "Canon tag c0 Unknown 11 (0x10c0 in raw)"},
   {0x00c1,  "Canon tag c1 Unknown 12 (0x10c1 in raw)"},
   {0x0100, "ImageWidth"},
   {0x0101, "ImageLength"},
   {0x0102, "Bits per sample"},
   {0x0103, "Compression"},
   {0x0106, "Photometric interpretation"},
   {0x010f, "Camera Make"},
   {0x0110, "Camera Model"},
   {0x0111, "Strips offsets"},
   {0x0112, "Orientation"},
   {0x0115, "Samples per pixel"},
   {0x0116, "Rows per strip"},
   {0x0117, "Strip byte counts"},
   {0x011A, "X REsolution"},
   {0x011B, "Y REsolution"},
   {0X011C, "Planar configuration"},
   {0x0128, "Resolution Unit"},
   {0x0132, "Date Time"},
   {0x0201, "JPEg Interchange format"},
   {0x0202, "JPEg Interchange format Length"},
   {0x0213, "YCbCrPositioning"},
   {0x829A, "Time of Exposure"},
   {0x829D, "F/Stop"},
   {0x8822, "Exposure PRogram???"},
   {0x8827, "ISO"},
   {0x9000, "Exif Version"},
   {0x9003, "DateTimeOriginal"},
   {0x9004, "DateTimeDigitized"},
   {0x9101, "ComponentsConfiguration"},
   {0x9102, "CompressedBitsPerPixel"},
   {0x9201, "ShutterSpeedValue"},  /* Units are in APEX = -log_2(exposuretime) */
   {0x9202, "ApertureValue"},      /* Units are in APEX = 2 log_2(F_Number) */
   {0x9204, "ExposureBiasValue"},  /* Units are in APEX */
   {0x9205, "MaxApertureValue"},
   {0x9206, "SubjectDistance"},
   {0x9207, "MeteringMode"},
   {0x9209, "Flash"},
   {0x920a, "FocalLength"},
   {TIFF_MAKER_NOTE_TAG, "MakerNote"},
   {0x9286, "UserComment"},
   {0xa000, "FlashpixVersion"},
   {0xa001, "ColorSpace"},
   {0xa002, "PixelXDimension"},
   {0xa003, "PixelYDimension"},
   {TIFF_INTEROPERABILITY_TAG, "InteroperabilityTag"},
   {0xa20e, "FocalPlaneXResolution"},
   {0xa20f, "FocalPlaneYResolution"},
   {0xa210, "FocalPlaneResolution Unit"},
   {0xa217, "Sensing method"},
   {0xa300, "File Source"},
   {0xa401, "Image CustomRendered"},
   {0xa402, "Exposure Mode"},
   {0xa403, "White balance"},
   {0xa406, "Scene Capture Type"},
   {TIFF_EXIF_TAG, "Exif IFD Offset"},


   {0xc5d8, "Unknown TIFF record d8"},
   {0xc5d9, "Unknown TIFF record d9"},
   {0xc5e0, "Unknown TIFF record e0"},

   {0,     NULL},
};

/* These are the basic TIFF format types */
#define BYTE_TIFF_TYPE 1
#define ASCII_TIFF_TYPE 2
#define SHORT_TIFF_TYPE 3
#define LONG_TIFF_TYPE 4
#define RATIONAL_TIFF_TYPE 5 //Two longs, first numerator, second denominator

/* Exif extends them with the following */

#define UNDEFINED_EXIF_FORMAT_TYPE 7 // Supposed to be array of bytes
#define SIGNED_RATIONAL_TIFF_TYPE 10 // Same as RATIONAL, but SIGNED

/* The size of each TIFF type */

static int tiffDataTypesSizes [] = {
   0,

   1,
   1,
   2,

   4,
   8,
   1,

   1,
   2,
   4,
   
   8,
   4,
   8
};



lookup_table_entry tiffTypeNames[] = {
   {BYTE_TIFF_TYPE, "BYTE"},
   {ASCII_TIFF_TYPE, "ASCII"},
   {SHORT_TIFF_TYPE, "SHORT"},
   {LONG_TIFF_TYPE, "LONG"},
   {RATIONAL_TIFF_TYPE, "RATIONAL"},
   {0, NULL},
};


char *Look_Up_Name(lookup_table_entry table[], int tag) 
{
   int i;
   i = 0;
   while (1) {
      if (table[i].name == NULL) {
        return "Unknown label!";
      }
      if (tag == table[i].id) {
        return table[i].name;
      }
      i++;
   }
}


class entry
{
public:
  entry() {;};
private:
  entry (const entry&);
  entry& operator= (const entry&);
};

class info_entry:public entry
{
protected:
  char *data;
  int length;
  entry *child;
public:
  info_entry() { data = NULL; length=-1;child = NULL;};
  virtual ~info_entry();
  void Set_Data(int _length, char *_data);
  void Set_Child(entry *_child) { child = _child; };
  entry* Get_Child() { return child; };
  int  Get_Data_as_Int();
  virtual int  Is_Container();
  virtual int  Get_Child_Offset() ;
  int Get_Length(void) { return length; };
  virtual void Print(int level);
  virtual void Print_Tag(int level);

};

int info_entry::Get_Child_Offset() 
{
  DEBUG_1("GEt child offset It should never be called!!!!\n");
  return 0;
  
}

int info_entry::Is_Container()
{
  return 0;
}

int info_entry::Get_Data_as_Int()
{
  int *value = (int*)data;
  int convertedValue = *value ;

  DEBUG_2("This is the value %d\n", convertedValue);

  return convertedValue;
}


void info_entry::Set_Data(int _length, char *_data)
{
  assert(_length < 16000);

  /* Let us do the byte order conversion here */

  length = _length;
  data = _data;
}

info_entry::~info_entry()
{
  DEBUG_1("Destructor of info_entry\n");
  if (data != NULL) {
    delete[] data;
  }
}

void info_entry::Print(int level)
{
  DEBUG_1("Printing\n");
}

void info_entry::Print_Tag(int level)
{
  DEBUG_1("Printing\n");
}


class tiff_info_entry:public info_entry
{

public:
  tiff_info_entry();
  ~tiff_info_entry();
  void Set(u_int16_t _id, u_int16_t _format, u_int16_t _num);
  int Size_In_Bytes(void);
  int Is_Container();
  int Get_Child_Offset() ;
  u_int16_t Get_Id(void) { return id; };
  void Print(int level);
  void Print_Tag(int level);
protected:
  u_int16_t id;
  u_int16_t format;  
  u_int32_t numberComponents; /*Number of values, not actual length, depends on "type"  */
  int len;
};

int tiff_info_entry::Is_Container()
{
  switch (Get_Id()) {
  case TIFF_EXIF_TAG:
    return 1;
  case TIFF_MAKER_NOTE_TAG:
    return 1;
  case TIFF_INTEROPERABILITY_TAG:
  default:
    return 0;
  }
}

int tiff_info_entry::Get_Child_Offset() 
{
  return Get_Data_as_Int();
}

void tiff_info_entry::Print_Tag(int level)
{
  char temp[100];
  memset(temp, ' ', level);
  temp[level] = 0;
  printf("%sRecord 0x%04x %s Length %d (0x%04x)\n", temp,id, Look_Up_Name(tiffTagsNames, id), len, len);
}




void tiff_info_entry::Print(int level)
{
  char temp[100];
  memset(temp, ' ', level);
  temp[level] = 0;
  printf("%sRecord 0x%04x %s Length %d (0x%04x)\n", temp,id, Look_Up_Name(tiffTagsNames, id), len, len);

  switch (format) {
  case LONG_TIFF_TYPE:
    {
      u_int32_t *pValue = (u_int32_t *) data;
      printf("%s    LONG Value:", temp);
      for (unsigned int i = 0; i<numberComponents; i++) {
	if (i >= 8) break;
	printf(" %12u", *(pValue + i));
      }
      printf("\n");
      printf("%s                  ", temp);
      for (unsigned int i = 0; i<numberComponents; i++) {
	if (i >= 8) break;
	printf(" x%08x", *(pValue + i));
      }
      printf("\n");
    }
    break;
  case ASCII_TIFF_TYPE:
    printf("%s    ASCII Value: [%s]\n", temp, data);
    break;
  case RATIONAL_TIFF_TYPE:
    {
      u_int32_t *pValue = (u_int32_t *) data;
      printf("%s      RATIONAL Value:", temp);
      for (unsigned int i = 0; i<numberComponents; i++) {
	if (i >= 8) break;
	printf(" %u/%u", *(pValue + i*2),*(pValue + i*2+1));
      }
      printf("\n");
      printf("%s                ", temp);
      for (unsigned int i = 0; i<numberComponents; i++) {
	if (i >= 8) break;
	if (*(pValue + i*2+1) != 0) {
	  printf(" %f", *(pValue + i*2) * 1.0 / *(pValue + i*2+1));
	} else {
	  printf(" INF ");
	}
      }
      printf("\n");
    }
    break;
  case SIGNED_RATIONAL_TIFF_TYPE:
    {
      int32_t *pValue = (int32_t *) data;
      printf("%s      SIGNED RATIONAL Value:", temp);
      for (unsigned int i = 0; i<numberComponents; i++) {
	if (i >= 8) break;
	printf(" %d/%d", *(pValue + i*2),*(pValue + i*2+1));
      }
      printf("\n");
      printf("%s                ", temp);
      for (unsigned int i = 0; i<numberComponents; i++) {
	if (i >= 8) break;
	if (*(pValue + i*2+1) != 0) {
	  printf(" %f", *(pValue + i*2) * 1.0 / *(pValue + i*2+1));
	} else {
	  printf(" INF ");
	}
      }
      printf("\n");
    }
    break;
  case SHORT_TIFF_TYPE:
    {
      u_int16_t *pValue = (u_int16_t *) data;
      printf("%s    SHORT Value:", temp);
      for (unsigned int i = 0; i<numberComponents; i++) {
	if (i >= 8) break;
	printf(" %5u", *(pValue + i));
      }
      printf("\n");
      printf("%s                ", temp);
      for (unsigned int i = 0; i<numberComponents; i++) {
	if (i >= 8) break;
	printf(" x%04x", *(pValue + i));
      }
      printf("\n");
    }
    break;
  case UNDEFINED_EXIF_FORMAT_TYPE:
  case BYTE_TIFF_TYPE:
    {
      unsigned char *pValue = (unsigned char *) data;
      if (format == UNDEFINED_EXIF_FORMAT_TYPE) 
	printf("%s    UNDEFINED TYPE Value:", temp);
      else
	printf("%s    SHORT Value:", temp);	  
      for (unsigned int i = 0; i<numberComponents; i++) {
	if (i >= 8) break;
	printf(" %3u", *(pValue + i));
      }
      printf("\n");
      printf("%s                ", temp);
      for (unsigned int i = 0; i<numberComponents; i++) {
	if (i >= 8) break;
	printf(" x%02x", *(pValue + i));
      }
      printf("\n");
    }
    break;
  default:
    fprintf(stderr, "TIFF format Type not supported %d\n", format);
    exit(1);
  }

}

tiff_info_entry::tiff_info_entry()
{
  //  DEBUG_1("Constructor for tiff_info_entry\n");
  id = 0;
  format = 0;
  len= 0;
  data = NULL;
}


void tiff_info_entry::Set(u_int16_t _id, u_int16_t _format, u_int16_t _num)
{


  id = _id;
  format = _format;
  numberComponents = _num;
  len = Size_In_Bytes();
  data = new char[len];
  assert(data != NULL);
}

tiff_info_entry::~tiff_info_entry()
{
  DEBUG_1("Destructor for tiff_info_entry\n");
  delete[] data;
}


int tiff_info_entry::Size_In_Bytes()
{
  int size;
  assert(format < sizeof(tiffDataTypesSizes)/sizeof(int));
	 
  size = tiffDataTypesSizes[format] * numberComponents;

  assert(size > 0);

  return size;
}


class dir_entry: public entry
{
protected:
  info_entry **entries; /* Array of pointers to the member entries of this directory */
  int count;            /* Number of member entries */
  dir_entry *pNext;     /* Next directory in the chain*/
public:
  int Get_Count() {return count;};
  virtual info_entry *Get_Entry(int i); /* Retrives the i-th member of the directory */
  dir_entry(int _count) {DEBUG_1("dir_entry contsturtor\n");count = _count; pNext = NULL;};
  void Set_Next(dir_entry *_pNext) {pNext = _pNext;};
  dir_entry* Get_Next(void) { return pNext; }; /* Returns next directory */
  virtual ~dir_entry();

};

dir_entry::~dir_entry()
{
  DEBUG_1("****Destructor dir_entry\n");
  for (int i=0;i<count;i++) {
    delete entries[i];
  }
  delete[] entries;
}

info_entry *dir_entry::Get_Entry(int i)
{
  return entries[i];
}


class tiff_dir_entry: public dir_entry
{
public:
  tiff_dir_entry(int _count);

  //  ~tiff_dir_entry();
  //  info_entry *Get_Entry(int i);
};


tiff_dir_entry::tiff_dir_entry(int _count) : dir_entry(_count)
{
  tiff_info_entry *pTemp;
  DEBUG_2("COnstructor tiff_dir_entry %d\n", count);
  count = _count; 
  assert(entries == NULL);
  DEBUG_1("To create entries\n");
  assert(count < 50);
  entries =  new info_entry*[count];
  for (int i=0;i<count;i++) {
    pTemp = new tiff_info_entry;
    entries[i] = (info_entry *)pTemp;
  }
}


class canon_parser 
{
public:
  //  canon_parser();
  canon_parser(FILE *file, byte_order bOrder);
  virtual dir_entry *Read_Next_Directory_Entry(void);
  virtual info_entry *Read_Next_Info_Entry(info_entry *entry);
  int Read_Next_Record(void);
  int Process_Record(void);
  virtual dir_entry *Parse(void);
  int Act_On_Record(int recordId);
  virtual int Init_Parser();
private:
  // Not defined anywhere
  canon_parser (const canon_parser&);
  canon_parser& operator= (const canon_parser&);
protected:
  FILE *input;
  byte_order byteOrder;
  info_entry *pEntries;
  int level;
};



canon_parser::canon_parser(FILE *file, byte_order bOrder)
{
  assert(file != NULL);
  input = file;
  byteOrder = bOrder;
  level = 0;
  DEBUG_2("Canon parser constructor %x\n", input);
}

/*

canon_parser::canon_parser()
{
  input = (FILE*)-1;
  DEBUG_1("Never to be execcuted\n");
}
*/
int canon_parser::Init_Parser()
{
  /* Do whatever is necessary to start parsing */

  DEBUG_1("THis is not meant to be called (init parser)!!!!!!!!!!!!!\n");
  return 0;
}

dir_entry *canon_parser::Read_Next_Directory_Entry()
{
  return NULL;
}

info_entry *canon_parser::Read_Next_Info_Entry(info_entry *entry)
{
  DEBUG_1("This should never be called!!!!!!!!!!!!!!!!!!!!\n");
  return entry;
}


dir_entry *canon_parser::Parse()
{
  int iCount=0;
  dir_entry *currentDir;
  dir_entry *previous = NULL;
  info_entry *info;
  dir_entry *first = NULL;
  /* 10 should be a decent value */

  ++level;
  assert(level < 10);
  DEBUG_1("************ To start parsing\n");

  while ((currentDir = Read_Next_Directory_Entry())  != NULL) {
    /* Maintain the link list of directories */
    if (previous != NULL) {
      previous->Set_Next(currentDir);
    } else {
      first = currentDir;
    }
    previous = currentDir;
    
    /* Now process each of its children */

    iCount = currentDir->Get_Count();
    DEBUG_2("We read %d entries in directory \n", iCount);

    for (int i=0;i<iCount;i++) {
      DEBUG_2("To get entry %d\n", i);
      info =currentDir->Get_Entry(i);
      DEBUG_2("CHeck Get Length: %d\n", info->Get_Length());
      Read_Next_Info_Entry(info);
#ifdef PRINT_WHILE_SCANNING
      info->Print(level); 
#endif

      if (info->Is_Container()) {
	int saveLocation = ftell (input);
	

	/* THis is a hack. To allow recursion we need to "unread" 4 bytes
	   so the first 4 bytes are the location of the next IFD */
	
	fseek(input, saveLocation - 4, SEEK_SET);

	info->Set_Child(Parse());

	DEBUG_1("************ End of parsing child\n");

	fseek(input, saveLocation, SEEK_SET);
      }

    }
    
  }
  --level;
  return first;
}

class canon_parser_tiff:public canon_parser
{
public:
  canon_parser_tiff(FILE *file, byte_order bOrder);

  info_entry *Read_Next_Info_Entry(info_entry *entry);
  dir_entry *Read_Next_Directory_Entry(void);

protected:
  int offsetEXIF;
  int entries;
  int insideMakerNote;
};



canon_parser_tiff::canon_parser_tiff(FILE *file, byte_order bOrder):
  canon_parser(file, bOrder)
{
  /* call  constructor of parent class */

  insideMakerNote = 0; /* This needs special treatment */
  offsetEXIF = 0; /* IN JPEGS, the EXIF are relative to the beginning of the
		     exif record, not the file.
		     in Canon CR2 they are relative to the beginning of the file */
  DEBUG_1("end of the constructor for Canon parser TIFF\n");

}



info_entry *canon_parser_tiff::Read_Next_Info_Entry(info_entry *entry)
{

  /* 
     This is what we need to read
     u_int16_t id;
     u_int16_t format;  
     u_int32_t numberComponents; 
     u_int32_t data;
   */
  u_int16_t id;
  u_int16_t format;  
  u_int32_t numberComponents; 
  u_int32_t value = 0;
  
  tiff_info_entry *pEntry;

  int iSize;
  char *data = NULL;

  pEntry = (tiff_info_entry*)entry;
  id = Read_Int_2(input, byteOrder);

  if (id == TIFF_MAKER_NOTE_TAG) {
    insideMakerNote = 1;
  }


  format = Read_Int_2(input, byteOrder);
  numberComponents = Read_Int_4(input, byteOrder);

  pEntry->Set(id, format, numberComponents);

  DEBUG_5("REading id %x format %d comp %d data %x \n", id, format, numberComponents, value);

  iSize = pEntry->Size_In_Bytes();

  
  /* Allocate the space for it */
  data = new char[iSize];

  if (iSize > 4) {
    DEBUG_2("TO read iSize %d\n", iSize);
    /* The data  is not here, it is somewhere else */

    value =  Read_Int_4(input, byteOrder);

    int saveLocation = ftell(input);

    
    /* Remmber, CR2s are offset to 0, but JPEGs are offset to beginning of EXIF */
    fseek(input, value+offsetEXIF, SEEK_SET);    
    
    /* Read as many bytes as we need */
    

    fread(data, 1, iSize, input);

    /* GO back to where we were */

    fseek(input, saveLocation, SEEK_SET);
  } else {
    /* The data is the value */
    fread(data, 1, 4, input);
  }
  /* now set it */

  DEBUG_6("REading id %x format %d comp %d data %x size %d\n", id, format, numberComponents, data, iSize);

  /* 
    We need to deal with the format and the endianism of the data

    The endianism matters for the following data types 

  */

  switch (format) {
  case SHORT_TIFF_TYPE :
    for (unsigned int i=0;i<numberComponents; i++) 
      Fix_Endianism_In_Record(data+i*2, "2", 2, byteOrder);
    break;
  case LONG_TIFF_TYPE :
    for (unsigned int i=0;i<numberComponents; i++) 
      Fix_Endianism_In_Record(data+i*4, "4", 4, byteOrder);
    break;
  case RATIONAL_TIFF_TYPE :
    for (unsigned int i=0;i<numberComponents; i++) 
      Fix_Endianism_In_Record(data+i*8, "44", 8, byteOrder);
    break;
  }
  if (format == ASCII_TIFF_TYPE) {
    DEBUG_2("Reading ascii %s\n", data);
  }
  


  pEntry->Set_Data(iSize, data);

  return entry;
}


class canon_parser_jpeg:public canon_parser_tiff
{
public:
  canon_parser_jpeg(FILE *file, byte_order bOrder);
  unsigned char Get_Next_Record_Type(void);
  int Init_Parser();

  /*
  info_entry *Read_Next_Info_Entry(info_entry *entry);
  virtual int Is_Container_Tag(info_entry *entry);
  */

};



dir_entry *canon_parser_tiff::Read_Next_Directory_Entry(void)
{
  DEBUG_1("We are to read next directory entry\n");

  int offsetIFD;
  offsetIFD = Read_Int_4(input, byteOrder);

  if (insideMakerNote && (offsetIFD == 0xffff || offsetIFD == 0)) {
    insideMakerNote = 0;
    return 0;
  }

  DEBUG_2("Offset read %x\n", offsetIFD);
  
  if (offsetIFD == 0) {
    return 0;
  }

  /* Remember, offsets are relative to EXIF record */
  offsetIFD += offsetEXIF;
  fseek(input, offsetIFD, SEEK_SET);



  entries = Read_Int_2(input, byteOrder);

#ifdef PRINT_WHILE_SCANNING
  printf("Reading new directory entry %d entries offset %d (%x)\n", entries, offsetIFD, offsetIFD);
#endif
  DEBUG_2("THere are %d entries\n", entries);


  return new tiff_dir_entry(entries);
}


canon_parser_jpeg::canon_parser_jpeg(FILE *file, byte_order bOrder):
  canon_parser_tiff(file, bOrder)
{
  /* call  constructor of parent class */
  DEBUG_1("end of the constructor for Canon parser jpeg\n");

}


unsigned char canon_parser_jpeg::Get_Next_Record_Type(void)
{
  unsigned char type;
  int i =8;
  while (--i) {
    type = fgetc(input);
    if (type != 0xff) 
      return type;
  }
  fprintf(stderr,"Too many markers\n");
  exit(1);
}

typedef struct {
  char exifLabel[6];       // 6
  u_int16_t byteOrder;     // 8
  u_int16_t exifMarker;   // 10
} exif_record;


int canon_parser_jpeg::Init_Parser(void)
{
  /* We need to eat all the data until we find the exif record */
  int length;
  int location;
  unsigned char marker;
  unsigned char recordType;
  exif_record exifHeader;
  int saveLocation;


  while (1) {
    /* Read new record */
    DEBUG_1("REading a record\n");
    location = ftell(input);
    marker = fgetc(input);

    assert(marker == 0xff);
    recordType = Get_Next_Record_Type();
    
    if (recordType == 0xd8) {
      break;
    }
    /* The lenght is always big endian */
    length = Read_Int_2(input, big_endian);
    /* Length contains includes the 2 bytes it takes to store itself */
    length -=2;

    DEBUG_2("Length %d\n", length);

    switch(recordType ) {
    case JPEG_EXIF:
      saveLocation = ftell(input);
      DEBUG_1("FOund exif\n");
      
      fread(&exifHeader, 1, sizeof(exifHeader), input);
      DEBUG_3("Label %s Endianism %x\n", exifHeader.exifLabel, exifHeader.byteOrder);
      

      if (strncmp(exifHeader.exifLabel, "Exif\0\0", 6) != 0) {
	/* It is not a exif! */
	fseek(input, saveLocation, SEEK_SET);
	/* Continue with another record */
	break;
      }
      /* Ok, so now we verify endianism */
      if (exifHeader.byteOrder == 0x4949) {
	byteOrder = little_endian;
      } else {
	byteOrder = big_endian;
      }
      
      {
	char *ptemp = (char*)& exifHeader;
	DEBUG_4("Offset byteOrder %d, maker %d, ifd %d \n",
		(char*)&(exifHeader.exifLabel) - ptemp, 
		(char*)&(exifHeader.byteOrder) - ptemp, 
		(char*)&(exifHeader.exifMarker) - ptemp
		);
      }		

      Fix_Endianism_In_Record(&(exifHeader.exifMarker), "2", 2, byteOrder);

      DEBUG_2("Marker 0x%02x\n", exifHeader.exifMarker);

      /* The offset is relative to the beginning of this record (after the exif identifier
	 (0x4949 are the first 2 bytes ) , not to the beginning of the file */
      offsetEXIF  += saveLocation + 6;

      DEBUG_2("Offset 0x%04x\n",offsetEXIF);

      return 1;
      break;
    case JPEG_SOS:
      /* the rest is the data */
      DEBUG_1("FOund sos\n");
      exit(1);
      return 0;
    default:
      fseek(input, length, SEEK_CUR);
    }
    
  }
  return 0;
}


class canon_parser_cr2:public canon_parser_tiff
{
public:
  canon_parser_cr2(FILE *file, byte_order bOrder);
  int Init_Parser();
};




int canon_parser_cr2::Init_Parser()
{
   
  /* The file pointer is pointing at the long integer with the location of the
     the first IFB */

  DEBUG_1("Inside Init parser cr2\n");

  DEBUG_2("Canon parser cr2 %x\n", input);


  return 1;
}

canon_parser_cr2::canon_parser_cr2(FILE *file, byte_order bOrder):
  canon_parser_tiff(file, bOrder)
{
  /* call  constructor of parent class */
  DEBUG_1("end of the constructor for Canon parser cr2\n");

}


class canon_photo
{
  FILE *input;
  byte_order byteOrder;
  canon_parser *parser;
public:
  canon_photo(FILE *file);
  ~canon_photo();
  void Parse();
  void Process_Directory();
protected:
  dir_entry *directory;
  void Process_Directory_Recursively(dir_entry *directory);
private:
  // Not defined anywhere
  canon_photo (const canon_photo&);
  canon_photo& operator= (const canon_photo&);
  int level;
};

canon_photo::canon_photo(FILE *file)
{

  char signature[4];

  directory = NULL;

  input = file;

  DEBUG_1("Starting\n");

   assert(file != NULL);
   /* We need to read the beginning of the file to determine the parser to use*/
   
  /* the first 4 characters tell us the type of file */
   fread(signature, 1, 4, input);
   
   if (strncmp(signature, "\x49\x49\x2a\00", 4) == 0) {
     byteOrder = little_endian;

     /* this is a CR2 file, with byte order intel */
     DEBUG_1("This is a CR2 file, about to call the constructor\n");
     parser = new canon_parser_cr2(file, byteOrder);
   } else if (strncmp(signature, "\xff\xd8\xff\xe1", 4) == 0) {
     /* move the file pointer to the 3rd byte, where the record starts */
     fseek(input, 2, SEEK_SET);
     DEBUG_1("This is a JPEG\n");
     parser = new canon_parser_jpeg(file, byteOrder);
   }
}


canon_photo::~canon_photo()
{
  /* I have to release the memory used by the entries */
  if (directory != NULL)
    delete directory;
  delete parser;
  fclose(input);
}

void canon_photo::Parse()
{
  if (parser->Init_Parser()) {
    directory = parser->Parse();
  }
}

void canon_photo::Process_Directory_Recursively(dir_entry *_directory)
{
  dir_entry *currentDir = _directory;
  info_entry *info;
  int iCount;

  level ++;
  debug = 1;
  
  //  DEBUG_1("TO start processing directory recursively\n");

  while (currentDir != NULL) {
    printf("Directory with %d entries\n", currentDir->Get_Count());

    iCount = currentDir->Get_Count();

    for (int i=0;i<iCount;i++) {
      info =currentDir->Get_Entry(i);
      info->Print_Tag(level);


      if (info->Is_Container()) {
	Process_Directory_Recursively((dir_entry*)info->Get_Child());
      }

    }



    currentDir = currentDir->Get_Next();
  }
  level --;
}


void canon_photo::Process_Directory()
{
  debug = 1;
  Process_Directory_Recursively(directory);
  level = 0;
}

void Verify_Packing(void)
{
  /* Verifies if the structs are "packed", that is, no space is in between 
     elements of a struct */
  if (sizeof(exif_record) != 10) {
    fprintf(stderr, "Please compile this program with -fpack-struct.\n");
  }
}


int main(int argc, char **argv)
{
  int arg;
  FILE *ifp;
  canon_photo *pPhoto;

  debug = 0;

  if (argc == 1) {
     fprintf (ERROR_OUTPUT, USAGE, argv[0]);
     exit(1);
  }

  for (arg=1; arg < argc; arg++) {
    printf("Processing %s\n", argv[arg]);
    ifp = fopen(argv[arg],"rb");
    if (!ifp) {
       perror(argv[arg]);
       continue;
    }
    if (debug) {
       fprintf (ERROR_OUTPUT, "Parsing %s:\n", argv[arg]);
    }
    pPhoto = new canon_photo(ifp);
    pPhoto->Parse();
    
    pPhoto->Process_Directory();

  }
  return 0;
}

