/* TODO

The names of the records depend on where we are:


+ IFD0 (main image)
  +SubIFD0
     Interoperability 
     Makernote 
+ IFD1 (thumbnail)
    

*/


#define DEBUG


#include <assert.h>
#include <assert.h>

#include <stdio.h>
#include <stdlib.h>
#include "libraw.h"
#include "crawdebug.h"


#define ENDIAN big_endian


byte_order byteOrder;



typedef struct {
   char exifLabel[6];
   u_int16_t byteOrder;
   u_int16_t exifMarker;
   u_int32_t offsetFirstIFD; /*14*/
   u_int32_t ifd0;
} exif_record;

typedef struct {
   u_int16_t numberEntries;
   u_int16_t dataEntry;
} ifd_record;

typedef struct {
   u_int16_t id;
   u_int16_t format;
   u_int32_t numberComponents;
   u_int32_t data;
} exif_entry;

enum data_types_exif  {unused0,
         unsigned_byte, ascii, unsigned_short, unsigned_long,
         unsigned_rational, 
         signed_byte, unused, signed_short, signed_long,
         signed_rational,
         signed_float, double_float};
         
typedef struct {
   int id;
   char *name;
} lookup_table_entry;



lookup_table_entry exifTagsNames[] = {
   {0x010f, "Make"},
   {0x0110, "Model"},
   {0x0112, "Orientation"},

   {0x011a, "X Resolution"},
   {0x011b, "Y Resolution"},

   {0x0128, "Resolution Unit"},
   {0x0132, "Date Time"},
   {0x0213, "YCbCrPositioning"},
   
   {0x8769, "Exif Offset"},

   {0x829a, "Time of Exposure"},
   {0x829d, "F/Stop"},
   {0x8827, "ISO"},
   {0x9000, "Exif Version"},
   {0x9003, "DateTimeOriginal"},
   {0x9004, "DateTimeDigitized"},
   {0x9101, "ComponentsConfiguration"},
   {0x9102, "CompressedBitsPerPixel"},
   {0x9201, "ShutterSpeedValue"},
   {0x9202, "ApertureValue"},
   {0x9204, "ExposureBiasValue"},
   {0x9205, "MaxApertureValue"},
   {0x9206, "SubjectDistance"},
   {0x9207, "MeteringMode"},
   {0x9209, "Flash"},
   {0x920a, "FocalLength"},
   {0x9286, "UserComment"},
   {0x927c, "MakerNote"},
   {0xa000, "FlashpixVersion"},
   {0xa001, "ColorSpace"},
   {0xa002, "PixelXDimension"},
   {0xa003, "PixelYDimension"},
   {0xa005, "InteroperabilityTag"},
   {0xa20e, "FocalPlaneXResolution"},
   {0xa20f, "FocalPlaneYResolution"},
   {0xa210, "FocalPlaneResolution Unit"},
   {0xa300, "File Source"},
   {0xa217, "Sensing method"},
   {0xa401, "Image CustomRendered"},
   {0xa402, "Exposure Mode"},
   {0xa403, "White balance"},
   {0xa406, "Scene Capture Type"},

   {0x0000,  "Canon tag 0 Always 0"},
   {0x0001,  "Canon tag 1 Photo Information (0x10d2 in raw first )"},
   {0x0002,  "Canon tag 2 Focal Length (0x5029 in raw)"},
   {0x0003,  "Canon tag 3 Unknown 1 (0x5028 in raw)"},
   {0x0004,  "Canon tag 4 Exposure information (0x102a in raw)"},
   {0x0006,  "Canon tag 6 Image name (0x0815 in raw)"},
   {0x0007,  "Canon tag 7 Camera Firmware (0x080b in raw)"},
   {0x0008,  "Canon tag 8  Photo serial number  (0x5817 in raw)"},
   {0x0009,  "Canon tag 9 Owner name (0x0810 in raw)"},

   {0x000c,  "Canon tag c Camera serial number (0x580b in raw)"},
   {0x000d,  "Canon tag d Image description??? (0x0805)"},
   {0x000e,  "Canon tag e"},
   {0x000f,  "Canon tag f Custom FNs           (0x1033 in raw)"},
   {0x0010,  "Canon tag 10"},
   {0x00c0,  "Canon tag c0 Unknown 11 (0x10c0 in raw)"},
   {0x00c1,  "Canon tag c1 Unknown 12 (0x10c1 in raw)"},
   {0x00aa,  "Canon tag aa Unknown 14 (0x10aa in raw)"},
   {0x00a8,  "Canon tag a8 Unknown 15 (0x10a8 in raw)"},
   {0x00a9,  "Canon tag a9 White balance table (0x10a9 in raw)"},

   {0x0103,  "Compression"},
   {0x0201,  "JPEGInterchangeFormat"},
   {0x0202,  "JPEGInterchangeFormatLength"},

   {0, NULL},
};

lookup_table_entry dataTypesNames[] = {
   {unused0,           "THis should not happen"},

   {unsigned_byte,     "unsigned byte"},
   {ascii,             "ascii"},

   {unsigned_short,    "unsigned short"},

   {unsigned_long,     "unsigned long"},
   {unsigned_rational, "unsigned rational"},
   
   {signed_byte,       "signed byte"},
   {unused,            "unused"},
   {signed_short,      "signed short"},
   {signed_long,       "signed long"},
   
   {signed_rational,   "signed rational"},
   
   {signed_float,      "signed float"},
   {double_float,      "double float"},
   {0, NULL},
};


#define SOF0  0xC0
#define SOF1  0xC1
#define SOF2  0xC2
#define SOF3  0xC3
#define SOF4  0xC4
#define SOF5  0xC5
#define SOF6  0xC6
#define SOF7  0xC7
#define SOF8  0xC8
#define SOF9  0xC9
#define SOF10 0xCA
#define SOF11 0xCB
#define SOF12 0xCC
#define SOF13 0xCD
#define SOF14 0xCE
#define SOF15 0xCF
#define SOI   0xD8
#define EOI   0xD9
#define SOS   0xDA
#define JFIF  0xE0
#define EXIF  0xE1
#define COM   0xFE

lookup_table_entry jpegRecordNames [] = {
  { SOF0,    "Start of frame"},
  { SOF1,    "Compression"},
  { SOF2,    "SOF2"},
  { SOF3,    "SOF3"},
  { SOF4,    "SOF4"},
  { SOF5,    "SOF5"},
  { SOF6,    "SOF6"},
  { SOF7,    "SOF7"},
  { SOF8,    "SOF8"},
  { SOF9,    "SOF9"},
  { SOF10,   "SOF10"},
  { SOF11,   "SOF11"},
  { SOF12,   "SOF12"},
  { SOF13,   "SOF13"},
  { SOF14,   "SOF14"},
  { SOF15,   "SOF15"},
  { SOI,     "Start of Image"},
  { EOI,     "End of Image"},
  { SOS,     "Start of Scan"},
  { JFIF,    "JFIF Marker"},
  { EXIF,    "Exif marker"},
  { COM,     "Comment"},
  { 0,       NULL},
};


lookup_table_entry orientationsNames[] = {
   {1,   "top left side"},
   {2,   "top right side"},
   {3,   "bottom  right side"},
   {4,   "bottom  left side"},
   {5,   "left side top"},
   {6,   "right side top"},
   {7,   "right side bottom"},
   {8,   "left side bottom"},
   {0, NULL}
};


char *Look_Up_Name(lookup_table_entry table[], int tag) 
{
   int i;
   i = 0;
   while (1) {
      if (table[i].name == NULL) {
        return "Unknown label!";
      }
      if (tag == table[i].id) {
        return table[i].name;
      }
      i++;
   }
}



int dataTypesSizes [] = {
   0,

   1,
   1,
   2,

   4,
   8,
   1,

   1,
   2,
   4,
   
   8,
   4,
   8
};

int outputBytesPerRow = 8;
char* outputPrefix = "";

void Print_Int2_Decimal(int length, void *data, int outputInt2PerRow, int byteOrder)
{
  int i=0;
  int int2Length = length / 2;
  int2 *int2Data = (int2 *)data;



  while (i<int2Length) {
    int j;

    printf("%03x (%d) %s    ", i, i, outputPrefix);
    for (j=0;j <outputInt2PerRow && i < int2Length;j++) {
       short int *value = (short int*)&(int2Data[i]);
       short int v;
       
       v = *value;

       printf("%6d %04x", v, v);
      i++;
    }
    printf("\n");
  }
}


void Print_Bytes_Hex(int length, void *data, int outputBytesPerRow)
{
  int i=0;

  while (i<length) {
    int j;
    printf("%03x %s    ", i,outputPrefix);
    for (j=0;j <outputBytesPerRow && i < length;j++) {
      unsigned char value = ((unsigned char*)data)[i];
      printf("x%02X ", value);
      if (value > 30 && value < 129) {
              printf("%c  ", value);
      } else {
              printf("   ");
      }
      i++;
    }
    printf("\n");
  }
}


int debug;


typedef struct {
   int16_t length;      //00   - 0 
   int16_t macroMode;   //02   - 2 -1 
   int16_t selfTimer;   //04   - 4 - 2/* Tenths of seconds */
   int16_t quality;     //06   - 
   int16_t flashMode;   //08
   int16_t continuousDrive;  //0a
   int16_t unknown1;         //0c
   int16_t focusingMode;     //0e
   int16_t unknown2;         //10
   int16_t captureMode;      //12 /* -1 NA, 0 unknown, 1 still/jpeg, 2 CCDRAW, 3movie*/
   int16_t imageSize;        //14
   int16_t easyShootingMode; //16
   int16_t digitalZoom;      //18
   int16_t imageContrast;    //1a
   int16_t imageSaturation;  //1c
   int16_t imageSharpness;   //1e
   int16_t imageISO;         //20  - 32 -16
   int16_t meteringMode;     //22    34
   int16_t focusingType;     //24    36
   int16_t autoFocusPointSelected;// 38
   int16_t exposureMode;       //    40
   int16_t unknown4;           //    42
   int16_t unknown5;           //    44
   int16_t lensFocalLengthMax; 
   int16_t lensFocalLengthMin; //30 - 48 - 24
   int16_t lensFocalLengthUnits;
   int16_t unknown6;  /* It is still unknown, 128 in all pictures*/
   int16_t unknown7;  /* SOme pictures it is 304, some 285, 320 */
   int16_t flashActivity;  //   - 56 - 28 /* 133 for external flash, 255 for internal */
   int16_t flashDetails;   //   - 58 - 29
                        /* 4008 for external flash, 2008 for internal flash */
                   /*Bits 15..0: (from David Burren)
                   14: external E-TTL
                   13: internal flash
                   11: FP sync used
                   7: 2nd("rear")-curtain sync used
                   4: FP sync enabled
                   other bits unknown
                   */
   u_int16_t u5;
   u_int16_t u6;
   u_int16_t u7;
   u_int16_t u8;
   u_int16_t u9;
   u_int16_t u10;
   u_int16_t u11;
   u_int16_t u12;
   u_int16_t u13;
   u_int16_t u14;
   u_int16_t u15;
   u_int16_t u16;
   u_int16_t u17;
   u_int16_t u18;
   u_int16_t u19;
   u_int16_t u20;
   u_int16_t u21;
   u_int16_t u22;
   u_int16_t u23;
   u_int16_t u24;
   u_int16_t u25;
   u_int16_t u26;
   u_int16_t u27;
   u_int16_t unknown[100];
} ciff_photo_info;

#define CIFF_PHOTO_INFO_SPEC  "2222222222" \
   "2222222222" \
   "2222222222" 

char *focusingModesNames[] = {
  "One shot",
  "AI Servo",
  "Unknown focusing mode",
  "Manual",
};

char *meteringModesNames[] = {
  "Unknown Metering Mode 0",
  "Unknown Metering Mode 1",
  "Unknown Metering Mode 2",
  "Evaluative",
  "Partial",
  "Centre",
};

char *exposureModesNames[] = {
  "Easy Mode (see Easy mode shooting)",
  "Program",
  "TV",
  "AV",
  "Manual",
  "A-Dept",
};

char *fileFormatsNames[] = {
  "",
  "JPEG",
  "Canon CRW",
};

char *compressionTypesNames[] = {
  "lossy JPEG",
  "none",
  "non-quantization JPEG",
};


char *targetImageTypesNames[] = {
  "Real-world subject",
  "Written document",
};


char *macroModesNames[] = {
  "",
  "Macro",
  "Normal",
};

char *flashModesNames [] = {
  "flash not fired",
  "auto",
  "on",
  "red-eye reduction",
  "slow synchro",
  "auto + red-eye reduction",
  "on + red-eye reduction",
  "",
  "",
  "",
  "",
  "",
  "",
  "",
  "",
  "",
  "external flash (not set on D30)",
};

char *imageSizesNames [] = {
  "Large",
  "Medium",
  "Small",
};


char *easyShootingModesNames [] = {
  "Full Auto",
  "None",
  "Landscape",
  "Fast Shutter",
  "Slow Shutter",
  "Night",
  "B&W",
  "Sepia",
  "Portrait",
  "Sports",
  "Macro / Close-Up",
  "Pan Focus",
};

char *captureModesNames [] = {
  "Unknown",
  "JPEG",
  "RAW",
  "Movie",
  "CM1",
  "CM2",
  "CM3",
  "CM4",
  "CM5",
  "CM6",
  "CM7",
  "CM8",
  "CM9",
  "CM10",
};




//----------------------------------------------------------------------
void Process_Record_Type_Photo_Info(unsigned short type, 
                                     ciff_photo_info *photoInfo)
{
  /* THis could be done better with an array... but that is for the future */

  
  assert(photoInfo->continuousDrive < sizeof(focusingModesNames)/sizeof(char*));
  assert(photoInfo->meteringMode < sizeof(meteringModesNames)/sizeof(char*));
  assert(photoInfo->exposureMode < sizeof(exposureModesNames)/sizeof(char*));
  assert(photoInfo->macroMode < sizeof(macroModesNames)/sizeof(char*));  
  assert(photoInfo->flashMode < sizeof(  flashModesNames)/sizeof(char*));  
  assert(photoInfo->imageSize < sizeof(  imageSizesNames)/sizeof(char*));  
  assert(photoInfo->easyShootingMode < sizeof( easyShootingModesNames)/sizeof(char*));  
  assert(photoInfo->captureMode < sizeof( captureModesNames)/sizeof(char*));  

  printf("%s   Macro Mode:               %s\n"
       "%s   Self Timer:               %4.2f seconds\n"
         "%s   Quality:                  %d \n"
         "%s   Flash Mode:               %s\n"
         "%s   Continuos Drive:          %d \n"
         "%s   Unknown1:                 %d\n"
         "%s   Focusing Mode:            %s\n" 
         "%s   Unknown2:                 %d\n"
         "%s   Capture mode:             %s\n"
         "%s   Image Size:               %s\n"
         "%s   Easy shooting mode:       %s\n"
         "%s   Digital Zoom:             %d\n"
         "%s   image contrast:           %d image saturation %d image sharpness %d\n"
         "%s   image ISO:                %d\n"
         "%s   Metering mode:            %s\n"
         "%s   Focusing Type:            %d\n"
         "%s   Autofocus point selected: 0x%x \n"
         "%s   exposure mode:            %s\n"
         "%s   unknown4                  0x%04x\n"
         "%s   unknown5                  0x%04x\n"
         "%s   unknown6                  0x%04x\n"
         "%s   unknown7                  0x%04x\n"
         "%s   flashActivity             0x%04x\n"
         "%s   flashDetails              0x%04x\n"

         ,
         outputPrefix,
         macroModesNames[photoInfo->macroMode],

         outputPrefix,
         photoInfo->selfTimer * 0.1,

         outputPrefix,
         photoInfo->quality,

         outputPrefix,
         flashModesNames[photoInfo->flashMode],
         
         outputPrefix,
         photoInfo->continuousDrive,

         outputPrefix,
         photoInfo->unknown1,

         outputPrefix,
         focusingModesNames[photoInfo->focusingMode],

         outputPrefix,
         photoInfo->unknown2,

         outputPrefix,
         captureModesNames[photoInfo->captureMode],
         
         outputPrefix,
         imageSizesNames[photoInfo->imageSize],

         outputPrefix,
         easyShootingModesNames[photoInfo->easyShootingMode],

         outputPrefix,
         photoInfo->digitalZoom,

         outputPrefix,
         photoInfo->imageContrast,
         photoInfo->imageSaturation,
         photoInfo->imageSharpness,

         outputPrefix,
         photoInfo->imageISO,

         outputPrefix,
         meteringModesNames[photoInfo->meteringMode],
         outputPrefix,

         photoInfo->focusingType,

         outputPrefix,
         photoInfo->autoFocusPointSelected,

         outputPrefix,
         exposureModesNames[photoInfo->exposureMode],
         
         outputPrefix,
         photoInfo->unknown4,
         outputPrefix,

         photoInfo->unknown5,

         outputPrefix,
         photoInfo->unknown6,
         outputPrefix,

         photoInfo->unknown7,
         outputPrefix,
         photoInfo->flashActivity,
         outputPrefix,
         photoInfo->flashDetails
         );

  if (photoInfo->lensFocalLengthMax == photoInfo->lensFocalLengthMin) {
    printf("%s   lens:                     %d mm\n",
           outputPrefix, photoInfo->lensFocalLengthMax * photoInfo->lensFocalLengthUnits);  

  } else {
    printf("%s   lens:                     %d-%d mm\n",
           outputPrefix, 
           photoInfo->lensFocalLengthMin * photoInfo->lensFocalLengthUnits,
           photoInfo->lensFocalLengthMax * photoInfo->lensFocalLengthUnits
           );  

  }


}
//----------------------------------------------------------------------

unsigned int Process_Data_Entries(char *exifData, char *data, int numberEntries, int recordOffset);


unsigned int Process_Exif_Entry(char *exifData, int nextOffset)
{
   ifd_record *ifd;
   unsigned int nextIFD = 0;

   nextOffset +=6;

   printf("Next offset %d %0x\n", nextOffset, nextOffset);

   ifd = (ifd_record*)((char*)exifData + nextOffset);

   Fix_Endianism_In_Record(ifd, "2", 4, byteOrder);

   //      printf("Number of entries %d \n", ifd->numberEntries);

   //   Print_Bytes_Hex(length, data);

   nextIFD = Process_Data_Entries(exifData, (char*)&(ifd->dataEntry), ifd->numberEntries, nextOffset);

   return nextIFD;
}


unsigned int Process_Data_Entries(char *exifData, char *data, int numberEntries, int recordOffset)
{
   int sizeBytes;
   int i=0;
   char *dataLocation;
   unsigned short int *value;
   u_int32_t nextIFD=0;

   Print_Bytes_Hex(12*numberEntries, data, 12);
   
   nextIFD = *(u_int32_t*)(12*numberEntries+data);
   Fix_Endianism_In_Record((char*)&nextIFD, "4", 4, byteOrder);
   

   Print_Bytes_Hex(4, 12*numberEntries+data, 12);



   for (i=0;i<numberEntries;i++) {

      exif_entry *exif = (exif_entry*)&data[i * 12];

      //      Print_Bytes_Hex(12, data+i*12);

      Fix_Endianism_In_Record((char*)exif, "224", 8, byteOrder);
      
      assert(exif->format <= 12);

      sizeBytes = dataTypesSizes[exif->format] *exif->numberComponents;


      
      printf("  %d exif id 0x%04x Tag name [%s]\n      format: %d [%s]   numberComp: %d Size in bytes%d\n",
             i+1,
             exif->id, Look_Up_Name(exifTagsNames, exif->id),
             exif->format, Look_Up_Name(dataTypesNames, exif->format),
             exif->numberComponents,
             sizeBytes);


      if (sizeBytes > 4) {
         Fix_Endianism_In_Record((char*)&exif->data, "4", 4, byteOrder);

         dataLocation = &data[exif->data-(recordOffset)+4];
      } else {
         dataLocation = (char*)&exif->data;
      }
      switch (exif->format) {
      case ascii:
         printf("     Ascii: [%s]\n", dataLocation);
         break;
      case unused:
         {
            if (exif->id == 0x927c) {
               unsigned int nextIFD;
               nextIFD = Process_Exif_Entry(exifData, exif->data);
               assert(nextIFD == 0);
               /*
               nextOffset = exif->data;
               printf("Next offset %d %x\n", nextOffset, nextOffset);
               */
            } else {
                        Print_Bytes_Hex(sizeBytes, dataLocation, 8);
            }
         }
         break;
      case unsigned_short:
         {
            ciff_photo_info* photoInfo ;

            if (exif->id == 0x01) {
               photoInfo = (ciff_photo_info*) dataLocation;

               //               Print_Bytes_Hex(sizeBytes, dataLocation, 2);               

               Fix_Endianism_In_Record((void*)photoInfo, CIFF_PHOTO_INFO_SPEC, sizeBytes, byteOrder);
               //               Print_Int2_Decimal(sizeBytes, dataLocation, 1, byteOrder);               

               Process_Record_Type_Photo_Info(0, photoInfo);
               //               Print_Int2_Decimal(sizeBytes, dataLocation, 1, byteOrder);               


            } else {
               value = (unsigned short int*)dataLocation;
               Fix_Endianism_In_Record(dataLocation, "2", 2, byteOrder);
               printf("   U Short: 0x%x %d\n", *value, *value);
            }
         }
         break;
      case unsigned_long:
         {
            u_int32_t *value;
            value = (u_int32_t *)dataLocation;
            Fix_Endianism_In_Record(dataLocation, "4", 4, byteOrder);
            printf("   U Long: 0x%x %d\n", *value, *value);
            if (exif->id == 0xa005 || exif->id == 0x8769) {
               unsigned int nextIFD;
               nextIFD = Process_Exif_Entry(exifData, exif->data);
               assert(nextIFD == 0);
               /*
               nextOffset = exif->data;
               printf("Next offset %d %x\n", nextOffset, nextOffset);
               */
            }
         }
         break;

      case unsigned_rational:
         {
            u_int32_t *value1;
            u_int32_t *value2;
            Fix_Endianism_In_Record(dataLocation, "44", 8, byteOrder);
            value1 = (u_int32_t *)dataLocation;
            value2 = (u_int32_t *)dataLocation + 1;
            printf("   Rational: %d/%d\n", *value2, *value1);
         }
         break;
      default:
         Print_Bytes_Hex(sizeBytes, dataLocation, 8);
         break;
      }
   }

   return nextIFD;
}


unsigned char Get_Record_Type(FILE *input)
{
  unsigned char type;
  int i =8;
  while (--i) {
    type = fgetc(input);
    if (type != 0xff) 
      return type;
  }
  fprintf(stderr,"Too many markers\n");
  exit(1);
}

int main(int argc, char *args[] )
{
   FILE *input;
   unsigned char  marker, recordType;
   int length;
   char *data=NULL;
   exif_record * exifData;
   int nextOffset;
   int count = 0;
        int location;
        int containsExif = 0;
   debug = 1;


   printf("Processing %d args %s\n", argc, args[1]);

   input = fopen(args[1], "rb");
   
   assert(input != NULL);

   marker = fgetc(input);


   recordType = Get_Record_Type(input);

   DEBUG_3("Record %x %s\n", recordType, Look_Up_Name(jpegRecordNames, recordType));

   assert(marker == 0xff && recordType == 0xd8);
   
   while (1) {

            /* Read new record */
          location = ftell(input);
          marker = fgetc(input);
          recordType = Get_Record_Type(input);
          
          if (recordType == 0xd8) {
                 break;
          }
          
          length = Read_Int_2(input, ENDIAN); 
          length -=2;
          
          DEBUG_5("Record %02x location %08x length %d %s\n", recordType, location, length, Look_Up_Name(jpegRecordNames, recordType));
          
          switch(recordType ) {
          case EXIF:
                 containsExif = 1;
                 length -=2;
                 data = malloc(length);
                 assert(data != NULL);
                 fread(data,  1, length, input);
                 break;
          case SOS:
                 /* the rest is the data */
                 return 0;
          default:
                 fseek(input, length, SEEK_CUR);
          }
   }
        /* Fix exit data */

   exifData = (exif_record*) data;
   /* We got to the JPEG data */
   printf("%s\n", exifData->exifLabel);
   printf("Byte order 0x%x\n", exifData->byteOrder);
        
        
   if (exifData->byteOrder == LITTLE_ENDIAN_ENCODING) {
          byteOrder = little_endian;
   } else {
          byteOrder = big_endian;
   }
        
   Fix_Endianism_In_Record(exifData, "cccccc" "cc24", 14, byteOrder);
        
   printf("Marker %x offset firstIFD %x\n", exifData->exifMarker, exifData->offsetFirstIFD);
        
   nextOffset = exifData->offsetFirstIFD;
        
   while (nextOffset>0) {
          switch(count++) {
      case 0:
		   printf("********IMAGE IFD\n");
         break;
      case 1:
		   printf("********THUMBNAIL IFD\n");
         break;
      default:
		   printf("********I DONT KNOW IFD\n");
         break;
      }
      

      nextOffset = Process_Exif_Entry((char*)exifData, nextOffset);
   }


   /* 14 */

   return 0;

}


