/*
   $Id: libciff.h,v 1.10 2005/03/24 06:12:15 dmg Exp $

   Library to read the records from a RAW file from Canon

   Based upon parse.c from David Coffin

   Credits:

     David Coffin:  Version 0.1 of this program is based on his parse.c
     David Burren:  Embedded exif record
     Peter Galbavy: His documentation of the format helped a lot


 * Copyright (C) 2004 Daniel M. German
 *
 * Author: Daniel M. German dmg _at_ turingmachine _dot_ org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */




#ifndef __LIBCIFF_H__
    #define __LIBCIFF_H__

#include "libcraw.h"

#define HEAP_CCDR_IDENT         "HEAPCCDR"
#define HEAP_CR2_IDENT          "\x00\x00" "CR" "\x02\x00" "\x12\xf2"

#ifndef __cplusplus

int ciff_Read_Data(craw_photo_type *pPhoto);

#else

//**********************************************************************

/* Definitions of the structs to properly interpret the
   records

See  Fix_Endianism_In_Record for an explanation of the SPEC records


 */

typedef struct {
   int            cameraNumber; 
   char           unknown[4];
} ciff_camera_serial_number;

#define CIFF_CAMERA_SERIAL_NUMBER_RECORD_SPEC  "4cccc"

//--------------------

typedef struct {
  unsigned short  unknown1; /* Zero in all SRLs?, non zero in G2 */
  unsigned short  focalLength;
  unsigned short  xSizeOfSensor; /* In 1000th inch */
  unsigned short  ySizeOfSensor; /* In 1000th inch */
  
  /* D60 
     50 mm lens EF 50 1.8 Mk I 
  Parsing /Users/dmg/photos/20040729/CRW_3153.CRW:
    | Focal length         0x5029 len        8 recOff  7262502 dataOff       -1
       Values as bytes
        x00 x00 x32 x00 x92 x03 x62 x02
       Values as bytes decimal
          0   0  50   0 146   3  98   2
       Values as int2
             0     50    914    610
       Values as int4
           3276800   39977874
  */

} ciff_focal_length;

#define CIFF_FOCAL_LENGTH_RECORD_SPEC  "2222"

//--------------------

typedef struct {
   int            recordLength;  
   unsigned short iso;   
   unsigned short unknown1; /* Looks like focal length FL-1 in 17-40, but in
			       28-105 it was 108 for 48, 52 for 90 */
   signed short targetAV;   // 10 bytes
   unsigned short targetTV; /* 5 */
  
   unsigned short exposureCompensation;
   unsigned short whiteBalance; // 16 bytes
   
   unsigned short unknown2[5];  
   unsigned short unknown3; /* 0 for flash off --> 0xffff for flash on*/
   
   unsigned short autoExposureCompensation;
   unsigned short flashExposureCompensation; /* 15 */
   
   unsigned short unknown4;
   
   unsigned short autoExposureBracketingCompensation;
   
   unsigned short unknown5[3];
   
   unsigned short actualAV;
   unsigned short actualTV;
   
   unsigned short unknown6[4];
   
} ciff_photo_exposure_information;

#define CIFF_PHOTO_EXPOSURE_INFORMATION_SPEC  "4222222" /* first chunck*/ \
    "222222"  /* 6 Unknowns  */\
    "22" /* autoex... flash... */\
    "2" /*unknown */\
    "2" /* autoexp... */\
    "222" /* 3 unknows */\
    "22"  /* actualAV, actualTV */\
    "2222" /* Last unknows */


/* THis one will hold values as we understand them */

typedef struct {
  int iso;

  double targetAV;
  double targetTV; /* 5 */
  
  int exposureCompensation;
  int whiteBalance;
  
  int autoExposureCompensation;
  int flashExposureCompensation; /* 15 */
  
  int autoExposureBracketingCompensation;
  
   double actualAV;
   double actualTV;
  
  
} ciff_translated_photo_exposure_information;

//--------------------

typedef struct {
   int index;
   int value;
} ciff_custom_function_entry;

typedef struct {
   int numberEntries;
   ciff_custom_function_entry entries[1]; /* THis will be allocated dynamically
					     to the actual size we need */
} ciff_translated_custom_functions;

//--------------------

typedef struct {
   unsigned short int  fileFormatCompression;
   float         targetCompressionRatio;
} ciff_image_format;

#define CIFF_IMAGE_FORMAT_RECORD_SPEC  "4f"

typedef struct {
   unsigned short int  fileFormat;
   unsigned short int  compressionType;
   float         targetCompressionRatio;
} ciff_translated_image_format;

//--------------------

typedef struct {
   unsigned int imageWidth;
   unsigned int imageHeight;
   float pixelAspectRatio;
   int   rotationAngle;
   unsigned int componentBitDepth;
   unsigned int colorBitDepth;
   unsigned int colorBW;
} ciff_image_details;

#define CIFF_IMAGE_DETAILS_RECORD_SPEC  "44f4444"


//--------------------

typedef struct {
   u_int16_t targetImageType;
   u_int16_t reserved1;
   u_int16_t reserved2;
   u_int16_t reserved3;
} ciff_target_image_type;

#define CIFF_TARGET_IMAGE_RECORD_SPEC  "2222"


//--------------------

typedef struct {
   u_int32_t serialNumber;
   u_int32_t reserved;
} ciff_translated_photo_serial_number;


#define CIFF_PHOTO_SERIAL_NUMBER_RECORD_SPEC  "44"

//--------------------


typedef struct {
  u_int32_t      dateTT;
  
  int32_t       timeZoneCode;
  u_int32_t      timeZoneInfo;

  /*
    TimeZoneInfo:

    Bit 31 is an indicator. 1 means that the ZoneCode is valid.
                            0 it is not used

    rest of bits are for future use

  */
} ciff_date;

#define CIFF_DATE_RECORD_SPEC  "444"
//--------------------


typedef struct {
   u_int16_t length;      //00   - 0 
   u_int16_t macroMode;   //02   - 2 -1 
   u_int16_t selfTimer;   //04   - 4 - 2/* Tenths of seconds */
   u_int16_t quality;     //06   - 
   u_int16_t flashMode;   //08
   u_int16_t continuousDrive;  //0a
   u_int16_t unknown1;         //0c
   u_int16_t focusingMode;     //0e
   u_int16_t unknown2;         //10
   u_int16_t captureMode;      //12 /* -1 NA, 0 unknown, 1 still/jpeg, 2 CCDRAW, 3movie*/
   u_int16_t imageSize;        //14
   u_int16_t easyShootingMode; //16
   u_int16_t digitalZoom;      //18
   u_int16_t imageContrast;    //1a
   u_int16_t imageSaturation;  //1c
   u_int16_t imageSharpness;   //1e
   u_int16_t imageISO;         //20  - 32 -16
   u_int16_t meteringMode;     //22
   u_int16_t focusingType;
   u_int16_t autoFocusPointSelected;
   u_int16_t exposureMode;
   u_int16_t unknown4;
   u_int16_t unknown5;
   u_int16_t lensFocalLengthMax; 
   u_int16_t lensFocalLengthMin; //30 - 48 - 24
   u_int16_t lensFocalLengthUnits;
   u_int16_t unknown6;  /* It is still unknown, 128 in all pictures*/
   u_int16_t unknown7;  /* SOme pictures it is 304, some 285, 320 */
   u_int16_t flashActivity;  //   - 56 - 28 /* 133 for external flash, 255 for internal */
   u_int16_t flashDetails;   //   - 58 - 29
                        /* 4008 for external flash, 2008 for internal flash */
		       /*Bits 15..0: (from David Burren)
			 14: external E-TTL
			 13: internal flash
			 11: FP sync used
			 7: 2nd("rear")-curtain sync used
			 4: FP sync enabled
			 other bits unknown
		       */
   u_int16_t u5;
   u_int16_t u6;
   u_int16_t u7;
   u_int16_t u8;
   u_int16_t u9;
   u_int16_t u10;
   u_int16_t u11;
   u_int16_t u12;
   u_int16_t u13;
   u_int16_t u14;
   u_int16_t u15;
   u_int16_t u16;
   u_int16_t u17;
   u_int16_t u18;
   u_int16_t u19;
   u_int16_t u20;
   u_int16_t u21;
   u_int16_t u22;
   u_int16_t u23;
   u_int16_t u24;
   u_int16_t u25;
   u_int16_t u26;
   u_int16_t u27;
   u_int16_t unknown[100];
} ciff_photo_info;

#define CIFF_PHOTO_INFO_SPEC  "2222222222" \
   "2222222222" \
   "2222222222" 

typedef struct {
   int length;
   int macroMode;
   int selfTimer; /* Tenths of seconds */
   int quality;
   int flashMode;
   int continuousDrive;
   int focusingMode;
   int captureMode; /* -1 NA, 0 unknown, 1 still/jpeg, 2 CCDRAW, 3movie*/
   int imageSize;
   int easyShootingMode;
   int digitalZoom;
   int imageContrast;
   int imageSaturation;
   int imageSharpness;
   int imageISO;
   int meteringMode;
   int focusingType;
   int autoFocusPointSelected;
   int exposureMode;
   int lensFocalLengthMax;
   int lensFocalLengthMin;
   int lensFocalLengthUnits;
   int flashInfo1;  /* It is still unknown, 128 in all pictures*/
   int flashInfo2;  /* SOme pictures it is 304, some 285 */
   int flashActivity; /* 133 for external flash, 255 for internal */
   int flashDetails;  /* 4008 for external flash, 2008 for internal flash */
	             /*Bits 15..0: (from David Burren)
			 14: external E-TTL
			 13: internal flash
			 11: FP sync used
			 7: 2nd("rear")-curtain sync used
			 4: FP sync enabled
			 other bits unknown
		       */
} ciff_translated_photo_info;


//**********************************************************************
class Ciff_Parser
{
private:
  craw_photo_type *pPhoto;
  int recursionLevel;
  char outputPrefix[256]; /* Used to indent output */
  int  outputBytesPerRow;
  int  outputInt2PerRow;
  int  outputInt4PerRow;
public:
  Ciff_Parser();
  void Set_Photo(craw_photo_type *pInitPhoto);
  int Parse(void);


private:
  char *Find_Record_Name(short type);

  void *Read_Data(unsigned int length, int offset);

  int   Ciff_Read_Records_In_Heap_Recursively(long dataOffset,
					      long lengthData);
  void Process_Record(unsigned short type, int length,
		      char *dataInHeader,
		      int recordOffset, int dataOffset);
  int Act_On_Record(unsigned short type);


  void Process_Record_Header(unsigned short type,
			     int length,
			     int recordOffset,
			     int dataOffset);

  void Process_Record_Type_Binary(unsigned short type,
				  int length,
				  void *data);

  void Process_Record_Type_String(unsigned short type, char *str);

  void Process_Record_Type_Date(unsigned short type,
				ciff_date *date);

  void Process_Record_Type_Camera_Serial_Number(unsigned short type,
					 ciff_camera_serial_number *serialNumberRecord);

  void Process_Record_Type_Photo_Serial_Number(unsigned short type,
					       ciff_translated_photo_serial_number *serialNumberRecord);


  void Process_Record_Type_Focal_Length(unsigned short type,
					ciff_focal_length *focalLength);

  void Process_Record_Type_Exposure_Info(unsigned short type,
					 ciff_translated_photo_exposure_information* exposure);

  void Process_Record_Custom_Functions(unsigned short type, 
				       ciff_translated_custom_functions *data);

  void Process_Record_Type_Photo_Info(unsigned short type, 
				      ciff_photo_info *photoInfo);

  void Process_Record_Type_Image_Format(unsigned short type, 
					ciff_translated_image_format *imageFormat);

  void Process_Record_Type_Image_Details(unsigned short type, 
					 ciff_image_details *imageDetails);

  void Process_Record_Type_Image_Details(unsigned short type, 
					 ciff_target_image_type *targetImage);


  void Print_Float4(int length, void *data);
  void Print_Int4_Decimal(int length, void *data);
  void Print_Int2_Decimal(int length, void *data);
  void Print_Bytes_Decimal(int length, void *data);
  void Print_Bytes_Hex(int length, void *data);


  void Translate_Exposure_Record(ciff_translated_photo_exposure_information *goodExposure,
				 ciff_photo_exposure_information *exposure);

  void Translate_Photo_Serial_Number_Record(ciff_translated_photo_serial_number *serialNumber,
				      void *data);

  void Translate_Custom_Functions(ciff_translated_custom_functions **pCustomFunctions,
				  void *data);

  void Translate_Image_Format(ciff_translated_image_format *imageFormat,
			      void *data);

};


#endif


#endif
