/*

   $Id: libcraw.h,v 1.8 2005/03/24 05:41:04 dmg Exp $

   Library to read the records from a RAW file from Canon

   Based upon parse.c from David Coffin

   Credits:

     David Coffin:  Version 0.1 of this program is based on his parse.c
     David Burren:  Embedded exif record 
     Peter Galbavy: His documentation of the format helped a lot


 * Copyright (C) 2004 Daniel M. German
 *
 * Author: Daniel M. German dmg _at_ turingmachine _dot_ org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#ifndef __LIBCRAW_H___
#define __LIBCRAW_H___


#include <stdio.h>
#include <stdlib.h>
#include "libraw.h"

/*
 * Header for use in the program
 */
typedef struct  {
   FILE          *file;
   byte_order     byteOrder;
   int            headerLength;
   unsigned char  identifier[9];
   long           fileLength;
   
   unsigned long           jpegOffset; /* LOcation of the embedded JPEG  in the file */
   unsigned long           jpegLength; /* Length of the embedded JPEG */
   
   char           manufacturerName[256];
   char           cameraName[256];
   unsigned long  cameraNumber;
   char           imageName[256];
   int            imageXSize;
   int            imageYSize;
   int            imageColorBits;
   int            imageFocalLength;
   int            imageSerialNumber;
   time_t         imageDate;
   int            macroMode;
   int            selfTimer;
   int            flashMode;
   int            continuousDrive;
   int            focusingMode;
   int            easyProgramMode;
   int            digitalZoom;
   int            imageContrast;
   int            imageSaturation;
   int            imageSharpness;
   int            imageSizeRelative;
   int            imageISO;
   int            imageAperture;
   int            imageTimeExposure;

   int            imageActualAperture;            /* The values need to be adjusted */
   int            imageActualTimeExposure;

   int            imageQuality;
   int            imageWhiteBalance;
   int            meteringMode;
   int            focusingType;
   int            autoFocusPointSelected;
   int            exposureMode;
   int            lensFocalLengthMin;
   int            lensFocalLengthMax;
   int            lensFocalLengthUnits;
   int            flashInfo1;
   int            flashInfo2;
   int            flashActivity;
   int            flashDetails;

   int            autoExposureBracketingCompensation;
   int            flashExposureCompensation; 


} craw_photo_type;

#define HEADER_LENGTH            32
#define HEADER_IDENTIFIER_LENGTH  8
#define HEADER_UNUSED_LENGTH     18


#ifdef __cplusplus
extern "C" {
#endif

craw_photo_type *craw_Load_Photo(FILE *photoFile);

void craw_Destroy_Photo(craw_photo_type *photo);

int craw_Get_Jpeg_Location(craw_photo_type *photo, 
			   unsigned long *pDataOffset, 
			   unsigned long *pDataLength);

#ifdef __cplusplus
}
#endif

#endif
