/*

   $Id: tiffdump.c,v 1.2 2005/03/24 06:12:15 dmg Exp $

   Scan the tiff file and dump each of its records. 

 * Copyright (C) 2004,2005 Daniel M. German
 *
 * Author: Daniel M. German dmg _at_ turingmachine _dot_ org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/* 

This program purpose is to demonstrate how to scan the records of a
TIFF file. TIFF defines a way in which an image and its metadata are
stored, not necessarily the format in which the image is actually
stored in it.

The CR2 files are TIFF files. 
*/

/*

SOme documentation with respect to the format, not very complete 

First 16 bytes:

4949  Endian type II
2a00  TIFF Version number, always the same
1000  pointer to first record (byte 16)

THe next 8 bytes are not used for data:

4352 0200 12f2 0800  CR


* Compression 6 corresponds to JPEG
* Tag 0x0201 JPEg Interchange format type points to beginning of JPEP



*/


/* TODO

The names of the records depend on where we are:


+ IFD0 (main image)
  +SubIFD0
     Interoperability 
     Makernote 
+ IFD1 (thumbnail)
    

*/


#define DEBUG


#include <assert.h>
#include <assert.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "libraw.h"
#include "crawdebug.h"


#define ENDIAN big_endian


byte_order byteOrder;



typedef struct {
   char exifLabel[6];
   u_int16_t byteOrder;
   u_int16_t exifMarker;
   u_int32_t offsetFirstIFD; /*14*/
   u_int32_t ifd0;
} exif_record;

typedef struct {
   u_int16_t numberEntries;
   u_int16_t dataEntry;
} ifd_record;

typedef struct {
   u_int16_t id;
   u_int16_t format;  
   u_int32_t numberComponents; /*Number of values, not actual length, depends on "type"  */
   u_int32_t data;
} tiff_data_entry;

enum data_types_exif  {unused0,
         unsigned_byte, ascii, unsigned_short, unsigned_long,
         unsigned_rational, 
         signed_byte, unused, signed_short, signed_long,
         signed_rational,
         signed_float, double_float};
         
typedef struct {
   int id;
   char *name;
} lookup_table_entry;


#define TIFF_EXIF_TAG             0x8769
#define TIFF_INTEROPERABILITY_TAG 0xa005
#define TIFF_MAKER_NOTE_TAG       0x927c

lookup_table_entry dataTypesNames[] = {
   {unused0,           "THis should not happen"},

   {unsigned_byte,     "unsigned byte"},
   {ascii,             "ascii"},

   {unsigned_short,    "unsigned short"},

   {unsigned_long,     "unsigned long"},
   {unsigned_rational, "unsigned rational"},
   
   {signed_byte,       "signed byte"},
   {unused,            "unused"},
   {signed_short,      "signed short"},
   {signed_long,       "signed long"},
   
   {signed_rational,   "signed rational"},
   
   {signed_float,      "signed float"},
   {double_float,      "double float"},
   {0, NULL},
};


#define SOF0  0xC0
#define SOF1  0xC1
#define SOF2  0xC2
#define SOF3  0xC3
#define SOF4  0xC4
#define SOF5  0xC5
#define SOF6  0xC6
#define SOF7  0xC7
#define SOF8  0xC8
#define SOF9  0xC9
#define SOF10 0xCA
#define SOF11 0xCB
#define SOF12 0xCC
#define SOF13 0xCD
#define SOF14 0xCE
#define SOF15 0xCF
#define SOI   0xD8
#define EOI   0xD9
#define SOS   0xDA
#define JFIF  0xE0
#define EXIF  0xE1
#define COM   0xFE

lookup_table_entry jpegRecordNames [] = {
  { SOF0,    "Start of frame"},
  { SOF1,    "Compression"},
  { SOF2,    "SOF2"},
  { SOF3,    "SOF3"},
  { SOF4,    "SOF4"},
  { SOF5,    "SOF5"},
  { SOF6,    "SOF6"},
  { SOF7,    "SOF7"},
  { SOF8,    "SOF8"},
  { SOF9,    "SOF9"},
  { SOF10,   "SOF10"},
  { SOF11,   "SOF11"},
  { SOF12,   "SOF12"},
  { SOF13,   "SOF13"},
  { SOF14,   "SOF14"},
  { SOF15,   "SOF15"},
  { SOI,     "Start of Image"},
  { EOI,     "End of Image"},
  { SOS,     "Start of Scan"},
  { JFIF,    "JFIF Marker"},
  { EXIF,    "Exif marker"},
  { COM,     "Comment"},
  { 0,       NULL},
};


lookup_table_entry orientationsNames[] = {
   {1,   "top left side"},
   {2,   "top right side"},
   {3,   "bottom  right side"},
   {4,   "bottom  left side"},
   {5,   "left side top"},
   {6,   "right side top"},
   {7,   "right side bottom"},
   {8,   "left side bottom"},
   {0, NULL}
};


lookup_table_entry tiffTagsNames [] = {
   {0x0000,  "Canon tag 0 Always 0"},
   {0x0001,  "Canon tag 1 Photo Information (0x10d2 in raw first )"},
   {0x0002,  "Canon tag 2 Focal Length (0x5029 in raw)"},
   {0x0003,  "Canon tag 3 Unknown 1 (0x5028 in raw)"},
   {0x0004,  "Canon tag 4 Exposure information (0x102a in raw)"},
   {0x0006,  "Canon tag 6 Image name (0x0815 in raw)"},
   {0x0007,  "Canon tag 7 Camera Firmware (0x080b in raw)"},
   {0x0008,  "Canon tag 8  Photo serial number  (0x5817 in raw)"},
   {0x0009,  "Canon tag 9 Owner name (0x0810 in raw)"},
   {0x000c,  "Canon tag c Camera serial number (0x580b in raw)"},
   {0x000d,  "Canon tag d Image description??? (0x0805)"},
   {0x000e,  "Canon tag e"},
   {0x000f,  "Canon tag f Custom FNs           (0x1033 in raw)"},
   {0x0010,  "Canon tag 10"},
   {0x00a8,  "Canon tag a8 Unknown 15 (0x10a8 in raw)"},
   {0x00a9,  "Canon tag a9 White balance table (0x10a9 in raw)"},
   {0x00aa,  "Canon tag aa Unknown 14 (0x10aa in raw)"},
   {0x00c0,  "Canon tag c0 Unknown 11 (0x10c0 in raw)"},
   {0x00c1,  "Canon tag c1 Unknown 12 (0x10c1 in raw)"},
   {0x0100, "ImageWidth"},
   {0x0101, "ImageLength"},
   {0x0102, "Bits per sample"},
   {0x0103, "Compression"},
   {0x0106, "Photometric interpretation"},
   {0x010f, "Camera Make"},
   {0x0110, "Camera Model"},
   {0x0111, "Strips offsets"},
   {0x0112, "Orientation"},
   {0x0115, "Samples per pixel"},
   {0x0116, "Rows per strip"},
   {0x0117, "Strip byte counts"},
   {0x011A, "X REsolution"},
   {0x011B, "Y REsolution"},
   {0X011C, "Planar configuration"},
   {0x0128, "Resolution Unit"},
   {0x0132, "Date Time"},
   {0x0201, "JPEg Interchange format"},
   {0x0202, "JPEg Interchange format Length"},
   {0x0213, "YCbCrPositioning"},
   {0x829A, "Time of Exposure"},
   {0x829D, "F/Stop"},
   {0x8822, "Exposure PRogram???"},
   {0x8827, "ISO"},
   {0x9000, "Exif Version"},
   {0x9003, "DateTimeOriginal"},
   {0x9004, "DateTimeDigitized"},
   {0x9101, "ComponentsConfiguration"},
   {0x9102, "CompressedBitsPerPixel"},
   {0x9201, "ShutterSpeedValue"},
   {0x9202, "ApertureValue"},
   {0x9204, "ExposureBiasValue"},
   {0x9205, "MaxApertureValue"},
   {0x9206, "SubjectDistance"},
   {0x9207, "MeteringMode"},
   {0x9209, "Flash"},
   {0x920a, "FocalLength"},
   {TIFF_MAKER_NOTE_TAG, "MakerNote"},
   {0x9286, "UserComment"},
   {0xa000, "FlashpixVersion"},
   {0xa001, "ColorSpace"},
   {0xa002, "PixelXDimension"},
   {0xa003, "PixelYDimension"},
   {TIFF_INTEROPERABILITY_TAG, "InteroperabilityTag"},
   {0xa20e, "FocalPlaneXResolution"},
   {0xa20f, "FocalPlaneYResolution"},
   {0xa210, "FocalPlaneResolution Unit"},
   {0xa217, "Sensing method"},
   {0xa300, "File Source"},
   {0xa401, "Image CustomRendered"},
   {0xa402, "Exposure Mode"},
   {0xa403, "White balance"},
   {0xa406, "Scene Capture Type"},
   {TIFF_EXIF_TAG, "Exif IFD Offset"},

   {0,     NULL},
};



char *Look_Up_Name(lookup_table_entry table[], int tag) 
{
   int i;
   i = 0;
   while (1) {
      if (table[i].name == NULL) {
        return "Unknown label!";
      }
      if (tag == table[i].id) {
        return table[i].name;
      }
      i++;
   }
}



int dataTypesSizes [] = {
   0,

   1,
   1,
   2,

   4,
   8,
   1,

   1,
   2,
   4,
   
   8,
   4,
   8
};

int outputBytesPerRow = 8;
char* outputPrefix = "";

void Print_Int2_Decimal(int length, void *data, int outputInt2PerRow, int byteOrder)
{
  int i=0;
  int int2Length = length / 2;
  int2 *int2Data = (int2 *)data;



  while (i<int2Length) {
    int j;

    printf("%03x (%d) %s    ", i, i, outputPrefix);
    for (j=0;j <outputInt2PerRow && i < int2Length;j++) {
       short int *value = (short int*)&(int2Data[i]);
       short int v;
       
       v = *value;

       printf("%6d %04x", v, v);
      i++;
    }
    printf("\n");
  }
}


void Print_Bytes_Hex(int length, void *data, int outputBytesPerRow)
{
  int i=0;

  while (i<length) {
    int j;
    printf("%03x %s    ", i,outputPrefix);
    for (j=0;j <outputBytesPerRow && i < length;j++) {
      unsigned char value = ((unsigned char*)data)[i];
      printf("x%02X ", value);
      if (value > 30 && value < 129) {
              printf("%c  ", value);
      } else {
              printf("   ");
      }
      i++;
    }
    printf("\n");
  }
}


int debug;


typedef struct {
   short int length;      //00   - 0 
   short int macroMode;   //02   - 2 -1 
   short int selfTimer;   //04   - 4 - 2/* Tenths of seconds */
   short int quality;     //06   - 
   short int flashMode;   //08
   short int continuousDrive;  //0a
   short int unknown1;         //0c
   short int focusingMode;     //0e
   short int unknown2;         //10
   short int captureMode;      //12 /* -1 NA, 0 unknown, 1 still/jpeg, 2 CCDRAW, 3movie*/
   short int imageSize;        //14
   short int easyShootingMode; //16
   short int digitalZoom;      //18
   short int imageContrast;    //1a
   short int imageSaturation;  //1c
   short int imageSharpness;   //1e
   short int imageISO;         //20  - 32 -16
   short int meteringMode;     //22    34
   short int focusingType;     //24    36
   short int autoFocusPointSelected;// 38
   short int exposureMode;       //    40
   short int unknown4;           //    42
   short int unknown5;           //    44
   short int lensFocalLengthMax; 
   short int lensFocalLengthMin; //30 - 48 - 24
   short int lensFocalLengthUnits;
   short int unknown6;  /* It is still unknown, 128 in all pictures*/
   short int unknown7;  /* SOme pictures it is 304, some 285, 320 */
   short int flashActivity;  //   - 56 - 28 /* 133 for external flash, 255 for internal */
   short int flashDetails;   //   - 58 - 29
                        /* 4008 for external flash, 2008 for internal flash */
                   /*Bits 15..0: (from David Burren)
                   14: external E-TTL
                   13: internal flash
                   11: FP sync used
                   7: 2nd("rear")-curtain sync used
                   4: FP sync enabled
                   other bits unknown
                   */
   unsigned short u5;
   unsigned short u6;
   unsigned short u7;
   unsigned short u8;
   unsigned short u9;
   unsigned short u10;
   unsigned short u11;
   unsigned short u12;
   unsigned short u13;
   unsigned short u14;
   unsigned short u15;
   unsigned short u16;
   unsigned short u17;
   unsigned short u18;
   unsigned short u19;
   unsigned short u20;
   unsigned short u21;
   unsigned short u22;
   unsigned short u23;
   unsigned short u24;
   unsigned short u25;
   unsigned short u26;
   unsigned short u27;
   unsigned short unknown[100];
} ciff_photo_info;

#define CIFF_PHOTO_INFO_SPEC  "2222222222" \
   "2222222222" \
   "2222222222" 

char *focusingModesNames[] = {
  "One shot",
  "AI Servo",
  "Unknown focusing mode",
  "Manual",
};

char *meteringModesNames[] = {
  "Unknown Metering Mode 0",
  "Unknown Metering Mode 1",
  "Unknown Metering Mode 2",
  "Evaluative",
  "Partial",
  "Centre",
};

char *exposureModesNames[] = {
  "Easy Mode (see Easy mode shooting)",
  "Program",
  "TV",
  "AV",
  "Manual",
  "A-Dept",
};

char *fileFormatsNames[] = {
  "",
  "JPEG",
  "Canon CRW",
};

char *compressionTypesNames[] = {
  "lossy JPEG",
  "none",
  "non-quantization JPEG",
};


char *targetImageTypesNames[] = {
  "Real-world subject",
  "Written document",
};


char *macroModesNames[] = {
  "",
  "Macro",
  "Normal",
};

char *flashModesNames [] = {
  "flash not fired",
  "auto",
  "on",
  "red-eye reduction",
  "slow synchro",
  "auto + red-eye reduction",
  "on + red-eye reduction",
  "",
  "",
  "",
  "",
  "",
  "",
  "",
  "",
  "",
  "external flash (not set on D30)",
};

char *imageSizesNames [] = {
  "Large",
  "Medium",
  "Small",
};


char *easyShootingModesNames [] = {
  "Full Auto",
  "None",
  "Landscape",
  "Fast Shutter",
  "Slow Shutter",
  "Night",
  "B&W",
  "Sepia",
  "Portrait",
  "Sports",
  "Macro / Close-Up",
  "Pan Focus",
};

char *captureModesNames [] = {
  "Unknown",
  "JPEG",
  "RAW",
  "Movie",
  "CM1",
  "CM2",
  "CM3",
  "CM4",
  "CM5",
  "CM6",
  "CM7",
  "CM8",
  "CM9",
  "CM10",
};




//----------------------------------------------------------------------
void Process_Record_Type_Photo_Info(unsigned short type, 
                                     ciff_photo_info *photoInfo)
{
  /* THis could be done better with an array... but that is for the future */

  
  assert(photoInfo->continuousDrive < sizeof(focusingModesNames)/sizeof(char*));
  assert(photoInfo->meteringMode < sizeof(meteringModesNames)/sizeof(char*));
  assert(photoInfo->exposureMode < sizeof(exposureModesNames)/sizeof(char*));
  assert(photoInfo->macroMode < sizeof(macroModesNames)/sizeof(char*));  
  assert(photoInfo->flashMode < sizeof(  flashModesNames)/sizeof(char*));  
  assert(photoInfo->imageSize < sizeof(  imageSizesNames)/sizeof(char*));  
  assert(photoInfo->easyShootingMode < sizeof( easyShootingModesNames)/sizeof(char*));  
  assert(photoInfo->captureMode < sizeof( captureModesNames)/sizeof(char*));  

  printf("%s   Macro Mode:               %s\n"
       "%s   Self Timer:               %4.2f seconds\n"
         "%s   Quality:                  %d \n"
         "%s   Flash Mode:               %s\n"
         "%s   Continuos Drive:          %d \n"
         "%s   Unknown1:                 %d\n"
         "%s   Focusing Mode:            %s\n" 
         "%s   Unknown2:                 %d\n"
         "%s   Capture mode:             %s\n"
         "%s   Image Size:               %s\n"
         "%s   Easy shooting mode:       %s\n"
         "%s   Digital Zoom:             %d\n"
         "%s   image contrast:           %d image saturation %d image sharpness %d\n"
         "%s   image ISO:                %d\n"
         "%s   Metering mode:            %s\n"
         "%s   Focusing Type:            %d\n"
         "%s   Autofocus point selected: 0x%x \n"
         "%s   exposure mode:            %s\n"
         "%s   unknown4                  0x%04x\n"
         "%s   unknown5                  0x%04x\n"
         "%s   unknown6                  0x%04x\n"
         "%s   unknown7                  0x%04x\n"
         "%s   flashActivity             0x%04x\n"
         "%s   flashDetails              0x%04x\n"

         ,
         outputPrefix,
         macroModesNames[photoInfo->macroMode],

         outputPrefix,
         photoInfo->selfTimer * 0.1,

         outputPrefix,
         photoInfo->quality,

         outputPrefix,
         flashModesNames[photoInfo->flashMode],
         
         outputPrefix,
         photoInfo->continuousDrive,

         outputPrefix,
         photoInfo->unknown1,

         outputPrefix,
         focusingModesNames[photoInfo->focusingMode],

         outputPrefix,
         photoInfo->unknown2,

         outputPrefix,
         captureModesNames[photoInfo->captureMode],
         
         outputPrefix,
         imageSizesNames[photoInfo->imageSize],

         outputPrefix,
         easyShootingModesNames[photoInfo->easyShootingMode],

         outputPrefix,
         photoInfo->digitalZoom,

         outputPrefix,
         photoInfo->imageContrast,
         photoInfo->imageSaturation,
         photoInfo->imageSharpness,

         outputPrefix,
         photoInfo->imageISO,

         outputPrefix,
         meteringModesNames[photoInfo->meteringMode],
         outputPrefix,

         photoInfo->focusingType,

         outputPrefix,
         photoInfo->autoFocusPointSelected,

         outputPrefix,
         exposureModesNames[photoInfo->exposureMode],
         
         outputPrefix,
         photoInfo->unknown4,
         outputPrefix,

         photoInfo->unknown5,

         outputPrefix,
         photoInfo->unknown6,
         outputPrefix,

         photoInfo->unknown7,
         outputPrefix,
         photoInfo->flashActivity,
         outputPrefix,
         photoInfo->flashDetails
         );

  if (photoInfo->lensFocalLengthMax == photoInfo->lensFocalLengthMin) {
    printf("%s   lens:                     %d mm\n",
           outputPrefix, photoInfo->lensFocalLengthMax * photoInfo->lensFocalLengthUnits);  

  } else {
    printf("%s   lens:                     %d-%d mm\n",
           outputPrefix, 
           photoInfo->lensFocalLengthMin * photoInfo->lensFocalLengthUnits,
           photoInfo->lensFocalLengthMax * photoInfo->lensFocalLengthUnits
           );  

  }


}
//----------------------------------------------------------------------

unsigned int Process_Data_Entries(char *exifData, char *data, int numberEntries, int recordOffset);


unsigned int Process_Exif_Entry(char *exifData, int nextOffset)
{
   ifd_record *ifd;
   unsigned int nextIFD = 0;

   nextOffset +=6;

   printf("Next offset %d %0x\n", nextOffset, nextOffset);

   ifd = (ifd_record*)((char*)exifData + nextOffset);

   Fix_Endianism_In_Record(ifd, "2", 4, byteOrder);

   //      printf("Number of entries %d \n", ifd->numberEntries);

   //   Print_Bytes_Hex(length, data);

   nextIFD = Process_Data_Entries(exifData, (char*)&(ifd->dataEntry), ifd->numberEntries, nextOffset);

   return nextIFD;
}


unsigned int Process_Data_Entries(char *exifData, char *data, int numberEntries, int recordOffset)
{
   int sizeBytes;
   int i=0;
   char *dataLocation;
   unsigned short int *value;
   u_int32_t nextIFD=0;

   Print_Bytes_Hex(12*numberEntries, data, 12);
   
   nextIFD = *(u_int32_t*)(12*numberEntries+data);
   Fix_Endianism_In_Record((char*)&nextIFD, "4", 4, byteOrder);
   

   Print_Bytes_Hex(4, 12*numberEntries+data, 12);



   for (i=0;i<numberEntries;i++) {

      tiff_data_entry *exif = (tiff_data_entry*)&data[i * 12];

      //      Print_Bytes_Hex(12, data+i*12);

      Fix_Endianism_In_Record((char*)exif, "224", 8, byteOrder);
      
      assert(exif->format <= 12);

      sizeBytes = dataTypesSizes[exif->format] *exif->numberComponents;


      
      printf("  %d exif id 0x%04x Tag name [%s]\n"
	     "          format: %d [%s]   numberComp: %d Size in bytes: %d value: %x\n",
             i+1,
             exif->id, Look_Up_Name(tiffTagsNames, exif->id),
             exif->format, Look_Up_Name(dataTypesNames, exif->format),
             exif->numberComponents,
             sizeBytes, exif->data
	     );


      if (sizeBytes > 4) {
         Fix_Endianism_In_Record((char*)&exif->data, "4", 4, byteOrder);

         dataLocation = &data[exif->data-(recordOffset)+4];
      } else {
         dataLocation = (char*)&exif->data;
      }
      switch (exif->format) {
      case ascii:
         printf("     Ascii: [%s]\n", dataLocation);
         break;
      case unused:
         {
            if (exif->id == 0x927c) {
               unsigned int nextIFD;
               nextIFD = Process_Exif_Entry(exifData, exif->data);
               assert(nextIFD == 0);
               /*
               nextOffset = exif->data;
               printf("Next offset %d %x\n", nextOffset, nextOffset);
               */
            } else {
	       Print_Bytes_Hex(sizeBytes, dataLocation, 8);
            }
         }
         break;
      case unsigned_short:
         {
            ciff_photo_info* photoInfo ;

            if (exif->id == 0x01) {
               photoInfo = (ciff_photo_info*) dataLocation;

               //               Print_Bytes_Hex(sizeBytes, dataLocation, 2);               

               Fix_Endianism_In_Record((void*)photoInfo, CIFF_PHOTO_INFO_SPEC, sizeBytes, byteOrder);
               //               Print_Int2_Decimal(sizeBytes, dataLocation, 1, byteOrder);               

               Process_Record_Type_Photo_Info(0, photoInfo);
               //               Print_Int2_Decimal(sizeBytes, dataLocation, 1, byteOrder);               


            } else {
               value = (unsigned short int*)dataLocation;
               Fix_Endianism_In_Record(dataLocation, "2", 2, byteOrder);
               printf("   U Short: 0x%x %d\n", *value, *value);
            }
         }
         break;
      case unsigned_long:
         {
            u_int32_t *value;
            value = (u_int32_t *)dataLocation;
            Fix_Endianism_In_Record(dataLocation, "4", 4, byteOrder);
            printf("   U Long: 0x%x %d\n", *value, *value);
            if (exif->id == 0xa005 || exif->id == TIFF_EXIF_TAG) {
               unsigned int nextIFD;
               nextIFD = Process_Exif_Entry(exifData, exif->data);
               assert(nextIFD == 0);
               /*
               nextOffset = exif->data;
               printf("Next offset %d %x\n", nextOffset, nextOffset);
               */
            }
         }
         break;

      case unsigned_rational:
         {
            u_int32_t *value1;
            u_int32_t *value2;
            Fix_Endianism_In_Record(dataLocation, "44", 8, byteOrder);
            value1 = (u_int32_t *)dataLocation;
            value2 = (u_int32_t *)dataLocation + 1;
            printf("   Rational: %d/%d\n", *value2, *value1);
         }
         break;
      default:
         Print_Bytes_Hex(sizeBytes, dataLocation, 8);
         break;
      }
   }

   return nextIFD;
}


unsigned char Get_Record_Type(FILE *input)
{
  unsigned char type;
  int i =8;
  while (--i) {
    type = fgetc(input);
    if (type != 0xff) 
      return type;
  }
  fprintf(stderr,"Too many markers\n");
  exit(1);
}


typedef struct {
   int2 byteOrder; /* MM Motorola ---big endian, II Intel --little endian*/
   int2 version;
   int4     offsetIFD;
} tiff_header;   

byte_order Byte_Order(int2 byteOrder) 
{
   u_int16_t *pByteOrder = (u_int16_t *)&byteOrder;
   if (*pByteOrder == 0x4949) {
      return little_endian;
   } else {
      return big_endian;
   }
}

#define BYTE_TIFF_TYPE 1
#define ASCII_TIFF_TYPE 2
#define SHORT_TIFF_TYPE 3
#define LONG_TIFF_TYPE 4
#define RATIONAL_TIFF_TYPE 5

lookup_table_entry tiffTypeNames[] = {
   {BYTE_TIFF_TYPE, "BYTE"},
   {ASCII_TIFF_TYPE, "ASCII"},
   {SHORT_TIFF_TYPE, "SHORT"},
   {LONG_TIFF_TYPE, "LONG"},
   {RATIONAL_TIFF_TYPE, "RATIONAL"},
   {0, NULL},
};



void Read_Directory_Entry(FILE *input, tiff_data_entry *tiffDE)
{
   fread(tiffDE, 1, 4, input);
   tiffDE->numberComponents = Read_Int_4(input, byteOrder); 
   tiffDE->data = Read_Int_4(input, byteOrder); 
}



int Process_Tiff_Directory(FILE *input, int offset)
{
   unsigned short int entries;
   int nextOffset ;
   static int level=-1;
   int i;
   tiff_data_entry tiffDir;
   int sizeBytes;
   char pad[60];
   int next;
   assert(++level < 10);

   memset(pad, ' ', level*3);
   pad[level*3] = 0;

   fseek(input, offset, SEEK_SET);
   entries = Read_Int_2(input, byteOrder);
   printf("%sTIFF Directory Entries [%d]\n", pad, entries);

   assert(sizeof(tiff_data_entry) == 12);
   for (i=0;i<entries;i++) {
      Read_Directory_Entry(input, &tiffDir);

      sizeBytes = dataTypesSizes[tiffDir.format] * tiffDir.numberComponents;

      printf("\n%s  Entry %d id 0x%04x Name [%s]\n"
	     "%s     format: %d [%s] numberComp: %d Size in bytes: %d Value: %04x (%u)\n",
	     pad,
             i+1,
             tiffDir.id, Look_Up_Name(tiffTagsNames, tiffDir.id),
	     pad,
             tiffDir.format, Look_Up_Name(dataTypesNames, tiffDir.format),
             tiffDir.numberComponents,
             sizeBytes,
	     (int)tiffDir.data,
	     (int)tiffDir.data
	     );

      switch (tiffDir.id) {
      case TIFF_EXIF_TAG:
      case TIFF_INTEROPERABILITY_TAG:
      case TIFF_MAKER_NOTE_TAG:
	 {
	    int saveLocation = ftell(input);
	    //	    assert(sizeBytes == 4);
	    next = Process_Tiff_Directory(input, (int)tiffDir.data);
	    fseek(input, saveLocation, SEEK_SET);
	    assert(next == 0);
	 }
	 break;
      }
   }
   nextOffset = Read_Int_4(input, byteOrder);
   --level;
   return nextOffset;

}
#ifdef REMOVED_CODE

      if (sizeBytes > 4) {
         Fix_Endianism_In_Record((char*)&exif->data, "4", 4, byteOrder);

         dataLocation = &data[exif->data-(recordOffset)+4];
      } else {
         dataLocation = (char*)&exif->data;
      }
      switch (exif->format) {
      case ascii:
         printf("     Ascii: [%s]\n", dataLocation);
         break;
      case unused:
         {
            if (exif->id == 0x927c) {
               unsigned int nextIFD;
               nextIFD = Process_Exif_Entry(exifData, exif->data);
               assert(nextIFD == 0);
               /*
               nextOffset = exif->data;
               printf("Next offset %d %x\n", nextOffset, nextOffset);
               */
            } else {
	       Print_Bytes_Hex(sizeBytes, dataLocation, 8);
            }
         }
         break;
      case unsigned_short:
         {
            ciff_photo_info* photoInfo ;

            if (exif->id == 0x01) {
               photoInfo = (ciff_photo_info*) dataLocation;

               //               Print_Bytes_Hex(sizeBytes, dataLocation, 2);               

               Fix_Endianism_In_Record((void*)photoInfo, CIFF_PHOTO_INFO_SPEC, sizeBytes, byteOrder);
               //               Print_Int2_Decimal(sizeBytes, dataLocation, 1, byteOrder);               

               Process_Record_Type_Photo_Info(0, photoInfo);
               //               Print_Int2_Decimal(sizeBytes, dataLocation, 1, byteOrder);               


            } else {
               value = (unsigned short int*)dataLocation;
               Fix_Endianism_In_Record(dataLocation, "2", 2, byteOrder);
               printf("   U Short: 0x%x %d\n", *value, *value);
            }
         }
         break;
      case unsigned_long:
         {
            u_int32_t *value;
            value = (u_int32_t *)dataLocation;
            Fix_Endianism_In_Record(dataLocation, "4", 4, byteOrder);
            printf("   U Long: 0x%x %d\n", *value, *value);
            if (exif->id == 0xa005 || exif->id == TIFF_EXIF_TAG) {
               unsigned int nextIFD;
               nextIFD = Process_Exif_Entry(exifData, exif->data);
               assert(nextIFD == 0);
               /*
               nextOffset = exif->data;
               printf("Next offset %d %x\n", nextOffset, nextOffset);
               */
            }
         }
         break;

      case unsigned_rational:
         {
            u_int32_t *value1;
            u_int32_t *value2;
            Fix_Endianism_In_Record(dataLocation, "44", 8, byteOrder);
            value1 = (u_int32_t *)dataLocation;
            value2 = (u_int32_t *)dataLocation + 1;
            printf("   Rational: %d/%d\n", *value2, *value1);
         }
         break;
      default:
         Print_Bytes_Hex(sizeBytes, dataLocation, 8);
         break;
      }
#endif






 int main(int argc, char *args[] )
 {
    FILE *input;
    tiff_header header;
    int offsetIFD;
    debug = 1;
    int j;
    int exifOffset;

    printf("Processing %d args %s\n", argc, args[1]);

    input = fopen(args[1], "rb");

    assert(input != NULL);

    fread(&header, 1, sizeof(tiff_header), input);

    /* One has to verify that the bytes 3 and 4 are 2a 00*/

    DEBUG_4("Header: [%x], version [%x], offset [%x] ", header.byteOrder.value, header.version.value,
	    (int)(header.offsetIFD.bytes));

    byteOrder = Byte_Order(header.byteOrder);

    offsetIFD = Convert_Number_4_bytes(header.offsetIFD, byteOrder);
    DEBUG_3("Offset First IFD: %d %x\n", offsetIFD, offsetIFD);

    while (offsetIFD != 0) {
       offsetIFD = Process_Tiff_Directory(input, offsetIFD);
    }

    exit(1);


    while (offsetIFD != 0) {
       int directoryEntryCount;
       tiff_data_entry directoryEntry;
       int i;
       /* Get to the begining of the record */
       fseek(input, offsetIFD, SEEK_SET);

       directoryEntryCount = Read_Int_2(input, byteOrder); 
       DEBUG_2 ("Number of entries in directory %d \n", directoryEntryCount);

       for ( i=0;i<directoryEntryCount;i++) {

	  //	 DEBUG_2("Reading entry %d\n", i);
	  Read_Directory_Entry(input, &directoryEntry);

	  DEBUG_5("Tag 0x%04x %-20s type %-10s length %05d ", directoryEntry.id,
		  Look_Up_Name(tiffTagsNames, directoryEntry.id),
		  Look_Up_Name(tiffTypeNames, directoryEntry.format), directoryEntry.numberComponents);

	  /*	 printf("Tag 0x%04x %s\n", directoryEntry.tag, Look_Up_Name(tiffTagsNames, directoryEntry.tag));
	   */

	  switch (directoryEntry.format) {
	  case BYTE_TIFF_TYPE :
	     break;
	  case ASCII_TIFF_TYPE :
	     if (directoryEntry.numberComponents > 4) {
		int saveLocation = ftell(input);
		char temp[100];
		assert(directoryEntry.numberComponents < 100);
		fseek(input, directoryEntry.data, SEEK_SET);
		fread(temp, 1, directoryEntry.numberComponents, input);
		DEBUG_1(temp);
		fseek(input, saveLocation, SEEK_SET);
	     } else {
		DEBUG_1("Not implemented\n");
		exit(1);
	     }
	     break;
	  case SHORT_TIFF_TYPE :
	     {
		unsigned short int temp;
		if (directoryEntry.numberComponents > 2) {
		   int saveLocation = ftell(input);
		   fseek(input, directoryEntry.data, SEEK_SET);
		   for (j=0;j<directoryEntry.numberComponents;j++) {
		      temp = Read_Int_2(input, byteOrder);
		      DEBUG_2("%d ", temp);
		   }
		   fseek(input, saveLocation, SEEK_SET);
		}  else {
		   u_int16_t *temp;
		   temp = (u_int16_t*)&directoryEntry.data;
		   Fix_Endianism_In_Record(&directoryEntry.data, "22", 4, byteOrder);
		   DEBUG_2("%d ", *temp);
		   if (directoryEntry.numberComponents == 2)
		      DEBUG_2("%d ", *(temp+1));
		}
	     }
	     break;
	  case LONG_TIFF_TYPE :
	     {
		unsigned int temp;
		if (directoryEntry.numberComponents > 1) {
		   int saveLocation = ftell(input);
		   fseek(input, directoryEntry.data, SEEK_SET);
		   for (j=0;j<directoryEntry.numberComponents;j++) {
		      temp = Read_Int_4(input, byteOrder);
		      DEBUG_3("0x%04x (%d) ", temp, temp);
		   }
		   fseek(input, saveLocation, SEEK_SET);
		}  else {
		   u_int32_t *temp;
		   temp = (u_int32_t*)&directoryEntry.data;
		   Fix_Endianism_In_Record(&directoryEntry.data, "4", 4, byteOrder);
		   DEBUG_3("0x%04x (%d) ", *temp, *temp);
		   if (directoryEntry.id == TIFF_EXIF_TAG) {
		      exifOffset = *temp;
		   }
		}
	     }
	     break;
	  case RATIONAL_TIFF_TYPE :
	     {
		int saveLocation = ftell(input);
		unsigned int denom, numer;
		fseek(input, directoryEntry.data, SEEK_SET);

		denom = Read_Int_4(input, byteOrder);
		numer = Read_Int_4(input, byteOrder);
		DEBUG_3("%d/%d ", numer, denom);
		fseek(input, saveLocation, SEEK_SET);

	     }
	     break;
	  default:
	     exit(1);
	  }
	  DEBUG_1("\n");

	  if (directoryEntry.id == TIFF_EXIF_TAG) {
	     /* This is the EXIF entry */
	     int saveLocation = ftell(input);
	     Process_Tiff_Directory(input, exifOffset);
	    fseek(input, saveLocation, SEEK_SET);
	 }
	 
      }
      offsetIFD = Read_Int_4(input, byteOrder); 

      DEBUG_3("Offset next IFD: %d %x\n", offsetIFD, offsetIFD);
   }
        /* Fix exit data */

   return 0;

}


